
#ifndef	DEFS_PMAIL_H
#define	DEFS_PMAIL_H


#include <time.h>
#include "message.h"
#include "mhead.h"
#include "libsupp.h"


#define	HOMEBOX		"mbox"
#define	OUTFOLDER	"postpone"
#define PAGER		"/bin/more"
#define	EDITOR		"/bin/vi"
#define TMPPATH		"/usr/tmp/mailzXXXXXX"
#define	PMTMP		"/usr/tmp/pmailXXXXXX"
#define DEFQUOTA	"Last time you wrote:\n"
#define	DELIMS		",:; "	/* delimiters for parsing RC file */


#define RCKEY_DEBUG	"log"		/* define log file */
#define RCKEY_DEFAULT	"default"	/* default operation */
#define RCKEY_VERBOSE	"verbose"	
#define RCKEY_TIMEOUT	"timeout"
#define RCKEY_SPAM	"spam"		/* incoming mail filter */
#define RCKEY_FETCH	"fetch"		/* fetch mail from incoming server */
#define RCKEY_TOPSLY	"topstyle"
//#define RCKEY_SIZELMT	"quota"		/* download message size quota */
#define RCKEY_LDELAY	"popdelay"	/* pop3 server login delay */
#define RCKEY_SMTP	"smtp"
#define RCKEY_SMBEHV	"smtpbeh"	/* default behavior of outgoing mail */
#define RCKEY_MBOX	"mbox"		/* default folder */
#define RCKEY_OUTBOX	"postpone"
#define RCKEY_BACKUP	"bakbox"
//#define RCKEY_SENDER	"sender"	/* act as ISP message sender */
#define RCKEY_REPLY	"reply-to"	/* act as From: */
#define RCKEY_SIG	"signature"
#define RCKEY_PIPE	"pipeto"	/* external pipe */
#define RCKEY_HEADER	"header"
#define RCKEY_PAGER	"pager"
#define RCKEY_EDITOR	"editor"
//#define RCKEY_DISHEAD	"showhead"	/* specify heads for displaying */
#define RCKEY_ADMEMO	"memo"		/* address memo file */
#define RCKEY_HEADMASK	"headmask"	/* Note 64 head classes limited */
 

#define COMMDELIM	'#'	/* start mark of commentory */
#define PATHDELIM	'/'	/* delimiter of directory */
#define MAX_DBG		16	/* specify the maximum debug and log device */
#define MAX_SIG		65535	/* the maximun size of signature file */
#define QUOTA_DBG	10240	/* default log file quota */
#define PATHDEP		32	/* maximum depth of relative file */
#define TASK_NR		32	/* maximum work classes limitation */
#define SVRBUF		256
#define EX_HEAD		128
#define MAXFOLDS	256
#define PARMAX  	100     /* maximum parameters in a command */
#define BLKBUF          4096


#define HMMAX   64              /* head mask limitation */
#define HMDEF   "From:To:Cc:Bcc:Subject:Date:Sender:Reply-To:X-Mailer"


#define	YES		1
#define	NO		0

#define	OK		0
#define WARNING		-1
#define	FAILURE		-2
#define	NOSMTPHOST	-3
#define	LOWMEMORY	-6
#define	OPENFILERR	-7
#define	CREATEFAIL	-8
#define	MISANSWER	-9
#define	LOGIN_FAIL	-10
#define SPAM_MSG	-20
#define	SPAM_FETCH	-21


#define LOG_ERR		9
#define LOG_INFO	2
#define	LOG_WARN	1
#define LOG_ALL		0	

#define	ID_NULL		0
#define	ID_GMAIL	1
#define	ID_STREAM	2
#define ID_HELP		3
#define ID_VERSION	4
#define	ID_TEMPLATE	5
#define	ID_DUMPTMP	6
#define	ID_DUMPINFO	7
#define	ID_SMTP		8
#define ID_HOSTLIST	9
#define ID_SPAMTEST	10
#define ID_PROGRESS	11

#define SMTP_PORT       25
#define POP3_PORT       110


/* this structure should be filled up according to 'smtp' option in the
 * config file. There are 4 entries in the 'smtp' option:
 *     host:reply_address:id:passwd
 * Pmail read the config file and chains all smtp structures. But only 
 * the first one, or the specified one, will be available. If there is
 * another smtp server specified in the command line, it should be the 
 * first item inserted to the chain. */

typedef struct _SMTP	{
	char	*host;
	char	*reply;
	char	*id;
	char	*passwd;
	struct	_SMTP *next;
	char	buf[1];
} SMTP;


#define	pfclose(ptr)  { \
  if ((ptr) != NULL)  fclose(ptr); \
  (ptr) = NULL; \
}

#define xfree(ptr)  {\
  free(ptr); \
  (ptr) = NULL; \
}

#define	chop(buf)	((buf)[strlen(buf)-1] = '\0')
 
#if	!defined(isspace)
#define	isspace(ch)	(((ch) > 0) && ((ch) <= ' '))
#endif

/* stupid strncpy cannot null-terminate oversized string, so I do it */

#define StrnCpy(d,s,l)	(strncpy((d),(s),(l)), (d)[(l)-1] = 0)


/* see main.c */

extern int  timeout;
extern int  verbose;
extern char *cmd_auth;
extern char *cmd_host;
extern char *cmd_mbox;
extern int  *cmd_cp_set;
extern int  cmd_action;
extern int  cmd_protocol;

int  act_bind(char *s);
void ReleaseResource();

/* see smtp.c */

int smtp_main(Folder *rootfld, SMTP *sm);
SMTP *smtp_host_chain(char *host, char *auth);
void smtp_host_list();
void smtp_host_clean();

/* see folder.c */

Folder *folder_open(char *mailbox, int beh);
Folder *folder_online(char *to, char *cc, char *bcc, char *sub, int beh);
int folder_close(char *backup, int beh);
int folder_close_online(char *backup, int beh);
int folder_number();
void folder_dump(Folder *fld);
void folder_list(Folder *fld);

/* mime.c */

char *decode_head(char *sour, int ctl);
int mime_body(Mail *m, FILE *fin, char *outfile, int ctl);

/* see strdate.c */

time_t strtimet(char *s);
int strdate(char *s, struct tm *t);
	
#endif

