
/*
   gmail.c -- the routine set of GMAIL protocol.
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include "defs.h"
#include "gmail.h"


static GMAIL	*gmail;
static struct	gm_att  *attr;

static GMAIL *gmail_create(char *s);
static int  gmail_set_protocal(char *s);
static Slnk *gmail_set_spam(char *mark);

#ifdef	DEBUG
static void gmail_dump_att();
static void gmail_dump(GMAIL *gm);
#endif

static Slnk *slnk_alloc();
static Slnk *spam_link_append(Slnk *anch, int id);
static int  sk_id(Spam *sc, char *p);
static int  sk_size(Spam *sc, char *p);
static int  sk_from(Spam *sc, char *p);
static int  sk_subject(Spam *sc, char *p);
static int  sk_action(Spam *sc, char *p);
	

int  gmail_single(GMAIL *gm, Head *head)
{
    int  rs = 0;
    
    switch (gm->protl) {
	case GP_POP3:
	case GP_APOP:
	    rs = pop3_main(gm, head);
	    break;
    }
    return rs;
}

int  gmail_main(GMAIL *gm, Head *head)
{
    //gmail_dump_att(); 
    while (gm)  {
        //gmail_dump(gm);
	gmail_single(gm, head);
	gm = gm->next;
    }
    return OK;
}

/* This function is used to create all necessary server objects for accessing
 * these servers. The server objects should be chained and start with 'gmail'.
 * Because there is a common structure gm_attr, it fills up this structure
 * first, then creates echo server object according to the 'fetch' option in
 * the configure file. All server objects will be chained at last with the 
 * order of first-appear-first-in. */

GMAIL *gmail_open()
{
    GMAIL *gm, *gtmp;
    struct winsize ws;
    char  *p, *idx[3], buf[SVRBUF];
    int   i, rs;

    /* if no default mailbox specified, we use 'mbox' in the current dir */
    
    if ((p = rc_check(RCKEY_MBOX)) == NULL)  
	strcpy(buf, HOMEBOX);
    else  
	fullpath(p, buf);

    /* There are three items which need extra space: cmd_auth, defbox and 
     * headmask. The plan is to combine the defbox and cmd_auth, then to 
     * append the headmask buffer at the tail of the combination. Therefore 
     * these is a pointer, named hdbuf, that points to the start position of 
     * the headmask buffer. */
    
    if (cmd_auth)  
	rs = strlen(cmd_auth) + strlen(buf);
    else  
	rs = strlen(buf);

    if ((p = rc_check(RCKEY_HEADMASK)) == NULL)
	rs += strlen(HMDEF);
    else
	rs += strlen(p);
    
    rs += sizeof(struct gm_att) + 5;

    if ((attr = malloc(rs)) == NULL)  
	return NULL;
    
    memset(attr, 0, rs);
    
    /* the extra space starts with 'buf' in the gm_att structure. we 
     * concatenate defbox and cmd_auth first and then split them */
    
    strcpy(attr->buf, buf);
    if (cmd_auth)  {
	strcat(attr->buf, ":");
	strcat(attr->buf, cmd_auth);
    }
    
    attr->hdbuf = attr->buf + strlen(attr->buf) + 1;
    if (p == NULL)
	strcpy(attr->hdbuf, HMDEF);
    else
	strcpy(attr->hdbuf, p);
    ziptoken(attr->hdbuf, attr->hdmsk, HMMAX, ":");
    
    rs = fixtoken(attr->buf, idx, 3, ":");
    for (i = 0; i < rs; i++)  {
	if (!strcmp(idx[i], "*"))  
	    idx[i] = NULL;
	else if (!strcmp(idx[i], "\\*"))  
	    idx[i]++;
    }
    attr->defbox = idx[0];
    attr->user = idx[1];
    attr->pass = idx[2];

    attr->act = cmd_action;
    attr->cp_set = cmd_cp_set;
    attr->mbox = cmd_mbox;

    
    if ((p = rc_check(RCKEY_TOPSLY)) != NULL)  {
	StrnCpy(buf, p, SVRBUF);
	ziptoken(buf, idx, 2, DELIMS);
	if (idx[0]) {
	    if (!strcmp(idx[0], "smart")) 
		attr->tops = TOP_SMART;
	    else if (!strcmp(idx[0], "all")) 
		attr->tops = TOP_ALL;
	    else if (!strcmp(idx[0], "brief"))
	        attr->tops = TOP_BRIEF;
	}
	if (idx[1])  
	    attr->topl = atonum(idx[1]);
	else  
	    attr->topl = 0;
    }

    if ((p = rc_check(RCKEY_LDELAY)) != NULL)  
	attr->delay = atoi(p);

    if (ioctl(1, TIOCGWINSZ, (char *) &ws) >= 0)  
	attr->cols = ws.ws_col;
    else {
	attr->cols = atoi(getenv("COLUMNS"));
	if (attr->cols == 0)  attr->cols = 80;
    }
    
    spam_queue_open();

    /* now creates the chain of server objects according to the relative
     * options in the configure file */
    
    for (p = rc_check(RCKEY_FETCH); p; p = rc_check_next())  {
	gm = gmail_create(p);
	if (gmail == NULL)  
	    gmail = gm;
	else {
	    for (gtmp = gmail; gtmp->next; gtmp = gtmp->next);
	    gtmp->next = gm;
	}
    }
    
    return gmail;
}

/* This function is used to insert an extra server object into the server
 * object chain. The argument, arg, comes from command line, usually by
 * -h option. Although it has the same form as the 'fetch' option in the 
 * configure file, the 'arg' often appears just a server address. 
 * NOTE that, if it is given with the form of "@number", such as "@2", 
 * it does not mean to create and insert a server object but to choose a
 * server object that has been created already in the chain with the series
 * number of 'number', in this case it's 2.  */
 
GMAIL  *gmail_cmdarg(char *arg)
{
    GMAIL *gm;
    int   i, rs;

    if (*arg == '@')  {
	rs = atoi(arg + 1);
	for (i = 1, gm = gmail; gm; i++, gm = gm->next)  {
	    if (i == rs) 
		return gm;
	}
    } else {		
	gm = gmail_create(arg);
	if (gm != NULL)  {
	    gm->next = gmail;
	    gmail = gm;
	}
    }
    
    return gm;
}
 
void gmail_close()
{
    GMAIL *gm, *tmp;

    for (gm = gmail; gm; )  {
	    
	free(gm->time_stamp);
	free(gm->mlist);
	
	if (gm->socket)  
	    close(gm->socket);
        if (gm->fp)  
	    fclose(gm->fp);

	tmp = gm;
	gm = gm->next;
	free(tmp);
    }
    spam_queue_free();
    
    if (attr)  
	free(attr);
    
    attr = NULL;
    gmail = NULL;
}


void gmail_host_list()
{
    GMAIL *gm;
    char  buf[SVRBUF];
    int   i;

    conPrintf("Post offices list:\n");
    
    for (i = 1, gm = gmail; gm; gm = gm->next, i++)  {
	ssprintf(buf, SVRBUF, "%4d %-25s %-25s %s:%s\n", 
			i, gm->host, gm->mbox, gm->lid, gm->pass);
	conPrintf(buf);
    }
    conPrintf("\n");
}

int gmail_set_action(char *s)
{
    int  rs = 0;
    
    switch (*s)  {
    case 'c': 
	break;
	      
    case 't':	/* for TOP */ 
    case 'l':
	rs |= GMAIL_LIST;
	break;
	
    case 'd':
	rs |= GMAIL_FETCH | GMAIL_DELE; 
	break;
	
    case 'f':
	rs |= GMAIL_FETCH;
	break;

    case 'r':
	rs |= GMAIL_DELE; 
	break;
    }
    return rs;
}

/* host.name:protocol:operation:spam:mailbox:id:passwd */

static GMAIL *gmail_create(char *s)
{
    GMAIL *gm;
    char  *idx[16];
    int   i, rs;

    if (s == NULL)  
	return NULL;
    
    /* becase the mailbox item is expanding to absolute path, I have to
     * increase another space about 256 bytes for it */
    
    rs = strlen(s) + sizeof(GMAIL) + PATHMAX + 2;
    if ((gm = malloc(rs)) == NULL)  
	return NULL;
    memset(gm, 0, rs);
    
    gm->attr = attr;
    gm->mbox = gm->buf;
    strcpy(gm->buf + PATHMAX + 1, s);
    
    rs = fixtoken(gm->buf + PATHMAX + 1, idx, 16, ":");
    for (i = 0; i < rs; i++)  {
	if (!strcmp(idx[i], "*"))  
	    idx[i] = NULL;
	else if (!strcmp(idx[i], "\\*"))  
	    idx[i]++;
    }
    
    if (cmd_host && *cmd_host != '@' && !isblankline(cmd_host))
	gm->host = cmd_host;
    else if (idx[0] && !isblankline(idx[0]))  
	gm->host = idx[0];
    else  {
	free(gm);
	return NULL;
    }
    
    gm->host = idx[0];
    
    /* specified the protocal to access server, POP3 is default */
    
    if (idx[1]) 
	gm->protl = gmail_set_protocal(idx[1]);
    if (gm->protl == 0)  
	gm->protl = GP_POP3;
    
    /* command line specified action could override the predefined one */
    
    if (idx[2]) 
	gm->act   = gmail_set_action(idx[2]);
    if (attr->act != ID_NULL)  
	gm->act = attr->act;
    
    if (idx[3]) 
	gm->slnk  = gmail_set_spam(idx[3]);
    gm->topspam = spam_need_top(gm->slnk);
    
    /* mailbox: command line specified > predefined > default */
    
    if (attr->mbox)  
	gm->mbox = attr->mbox;	/* should be expanded */
    else if (idx[4] && !isblankline(idx[4]))  
	fullpath(idx[4], gm->mbox);
    else 
	gm->mbox = attr->defbox;
    
    if (attr->user)  
	gm->lid = attr->user;
    else  
	gm->lid  = idx[5];

    if (attr->pass)  
	gm->pass = attr->pass;
    else  
	gm->pass = idx[6];
    
    return gm;
}

static int gmail_set_protocal(char *s)
{
    if (!strcasecmp(s, "pop3"))       
	return GP_POP3;
    else if (!strcasecmp(s, "apop"))  
	return GP_APOP;
    return 0;
}

static Slnk *gmail_set_spam(char *mark)
{
    Slnk  *anch = NULL;
    char  buf[128], *idx[32];	/* this buffer might be overflow attacked */
    int   i, id;

    if ((*mark == 'y') || (*mark == 'Y'))  {
	anch = spam_link_append(anch, 0);
    } else if ((*mark != 'n') && (*mark != 'N'))  {
	StrnCpy(buf, mark, sizeof(buf));
	ziptoken(buf, idx, 32, ",");
	for (i = 0; idx[i]; i++)  {
	    id = atoi(idx[i]);
	    anch = spam_link_append(anch, id);
	}
    }
    return anch;
}

/********************** process spam routine ********************/

static Spam  *spam;

static Slnk  slnk[SLPOOL];
static int   slidx = 0;


static Slnk  *slnk_alloc()
{
    if (slidx >= SLPOOL)  return NULL;
    slidx++;
    return &slnk[slidx - 1];
}

/*
static char *strcasestr(const char *haystack, const char *needle)
{
    char  *p;
    int   nlen;

    if (!needle || !haystack)  return (char*) haystack;

    nlen = strlen(needle);
    for (p = (char*) haystack; *p; p++)  {
        if (!strncasecmp(p, (char *) needle, nlen))  return p;
    }
    return NULL;
}
*/

static Slnk *spam_link_append(Slnk *anch, int id)
{
    Spam  *sc;
    Slnk  *sl;

    for (sc = spam; sc; sc = sc->next)  {
	if (sc->id == id)  {
	    if ((sl = slnk_alloc()) == NULL)  break;
	    sl->sc = sc;
	    sl->next = anch;
	    anch = sl;
	}
    }
    return anch;
}

/*****************************************************************
 * spam_check_xxxx()  and spam_regex_xxxx()
 * return:
 * 0: nothing matched; 
 * 1: matched something
 **/

static char *allfrom[] = { "from", "reply", "sender", NULL };

/*
static int spam_check_from(Head *head, char *from)
{
    struct head_chain  *hc;
    HLin   *hl;
    int    i, rs;
    
    for (i = rs = 0; allfrom[i]; i++)  {
	    
	if ((hl = head_check(head, allfrom[i])) != NULL)  {
	    if (strcasestr(hl->body, from))  rs++;
	    for (hc = hl->hc; !rs && hc; hc = hc->next)  {
		if (strcasestr(hc->str, from))  rs++;
	    }
	}
	
	if (rs)  break;
    }
    return rs;
}

static int spam_check_subject(Head *head, char *subject)
{
    struct head_chain  *hc;
    HLin   *hl;
    int    rs = 0;
    
    if ((hl = head_check(head, "subject")) != NULL)  {
	if (strcasestr(hl->body, subject))  rs++;
	for (hc = hl->hc; !rs && hc; hc = hc->next)  {
	    if (strcasestr(hc->str, subject))  rs++;
	}
    }
    return rs;
}
*/


static int spam_regex_from(Head *head, regex_t *from)
{
    regmatch_t  pmatch[1];
    struct head_chain  *hc;
    HLin   *hl;
    int    i, rs;
    
    for (i = rs = 0; allfrom[i]; i++)  {
	    
	if ((hl = head_check(head, allfrom[i])) != NULL)  {
	    if (!regexec(from, hl->body, 1, pmatch, 0))  rs++;
	    for (hc = hl->hc; !rs && hc; hc = hc->next)  {
		if (!regexec(from, hc->str, 1, pmatch, 0))  rs++;
	    }
	}
	
	if (rs)  break;
    }
    return rs;
}

static int spam_regex_subject(Head *head, regex_t *subject)
{
    regmatch_t  pmatch[1];
    struct head_chain  *hc;
    HLin   *hl;
    int    rs = 0;
    
    if ((hl = head_check(head, "subject")) != NULL)  {
	if (!regexec(subject, hl->body, 1, pmatch, 0))  rs++;
	for (hc = hl->hc; !rs && hc; hc = hc->next)  {
	    if (!regexec(subject, hc->str, 1, pmatch, 0))  rs++;
	}
    }
    return rs;
}


/* 1 spam message, 0 good message */

static int  spam_cmp(Spam *sc, Head *head, int msize)
{
    if (!sc || !head)  return 0;
    
    if (sc->quota && (msize < sc->quota))  goto trigr;

    if (sc->from_reg)  {
	if (spam_regex_from(head, sc->from_reg) == 0)  goto trigr;
    }
    if (sc->sub_reg)  {
	if (spam_regex_subject(head, sc->sub_reg) == 0)  goto trigr;
    }

    /* if acknowleged all tests ... */

    if (sc->logic == 0)  return 1;
    else return 0;

trigr:		/* any one unmatch */
    return sc->logic;
}

Spam *spam_match_chain(Slnk *sl, Head *head, int msize)
{
    while (sl)  {
	if (spam_cmp(sl->sc, head, msize))  return sl->sc;
	sl = sl->next;
    }
    return NULL;
}

/* this funtion is used to check spam mails by their size only */

Spam *spam_match_size(Slnk *sl, int size)
{
    Spam  *sc;
    
    while (sl)  {
	sc = sl->sc;
	if (!sc->quota || sc->from_reg || sc->sub_reg)  continue;
	
	if (size > sc->quota) {
	    if (sc->logic == 0)
		return sc;
	} else {
	    if (sc->logic != 0)
		return sc;
	} 
	
	sl = sl->next;
    }
    return NULL;
}	

int  spam_need_top(Slnk *sl)
{
    int  rs = 0;

    while (sl)  {
	if (sl->sc->from_reg || sl->sc->sub_reg)  rs++;
	sl = sl->next;
    }
    return rs;
}
    

/**********************************************************************/

static int sk_id(Spam *sc, char *p)
{
    sc->id = atoi(p);
    return 0;	/* this is identity, not an entry. so count free */
}

static int sk_size(Spam *sc, char *p)
{
    sc->quota = atonum(p);
    return 1;
}

static int sk_from(Spam *sc, char *p)
{
    if (isblankline(p))  return 0;
    sc->from = dup_str(p);
    
    sc->from_reg = &sc->preg[sc->ireg];
    if (regcomp(sc->from_reg, p, REG_EXTENDED | REG_ICASE))  {
	dbgPrintf("sk_from: cannot compile filter pattern.\n");
	sc->from_reg = NULL;
    } else sc->ireg++;

    return 1;
}
	
static int sk_subject(Spam *sc, char *p)
{
    if (isblankline(p))  return 0;
    sc->subject = dup_str(p);
    
    sc->sub_reg = &sc->preg[sc->ireg];
    if (regcomp(sc->sub_reg, p, REG_EXTENDED | REG_ICASE))  {
	dbgPrintf("sk_subject: cannot compile filter pattern.\n");
	sc->sub_reg = NULL;
    } else sc->ireg++;

    return 1;
}

static int sk_action(Spam *sc, char *p)
{
    sc->logic = 0;
    if (!strcasecmp(p, SA_YES_DELE))      
	sc->action = SPAM_DELE;
    else if (!strcasecmp(p, SA_YES_KEEP)) 
	sc->action = SPAM_KEEP;
    else if (!strcasecmp(p, SA_YES_ASK))  
	sc->action = SPAM_ASK;
    else if (!strcasecmp(p, SA_NO_DELE))  
	sc->action = SPAM_DELE, sc->logic = 1;
    else if (!strcasecmp(p, SA_NO_KEEP))  
	sc->action = SPAM_KEEP, sc->logic = 1;
    else if (!strcasecmp(p, SA_NO_ASK))   
	sc->action = SPAM_ASK, sc->logic = 1;
    return 0;	/* action is count free either */
}
	
static struct	spam_func	{
	
	char	*key;
	int	(*func)(Spam *sc, char *p);	/* successed: 1, fail: 0 */
	
} sf[] = {
	{ SK_ID,	sk_id },
	{ SK_SIZE,	sk_size },
	{ SK_FROM,	sk_from },
	{ SK_SUBJECT,	sk_subject },
	{ SK_ACTION, 	sk_action },
	{ NULL, NULL }
};
	

Spam *spam_queue_open()
{
    Spam  *sc;
    char  *p, *s, *tmp = NULL, *idx[16];
    int   rs, i, k, cnt;

#ifdef	DEBUG
    if (spam)  {
	printf("spam queue has existed.\n");
	return NULL;
    }
#endif

    for (p = rc_check(RCKEY_SPAM); p; p = rc_check_next())  {
	
	p += strlen(p) + 1;	/* move to extension part */
	//puts(p);
	
	tmp = dup_str(p);
	if (tmp == NULL)  break;
	
	rs = ziptoken(tmp, idx, 16, "#"); 

	if ((sc = malloc(sizeof(Spam))) == NULL)  {
	    free(tmp);
	    break;
	}
	memset(sc, 0, sizeof(Spam));

	for (cnt = 0, i = 1; i < rs - 1; i++)  {
	    for (k = 0; sf[k].key; k++)  {
		if (!strncasecmp(idx[i], sf[k].key, strlen(sf[k].key))) {
		    s = skip_space(idx[i] + strlen(sf[k].key));
		    cnt += sf[k].func(sc, s);
		    break;
		}
	    }
	}

	if (cnt == 0)  free(sc);
	else if (spam == NULL)  spam = sc;
	else {
	    sc->next = spam;
	    spam = sc;
	}
	free(tmp);
    }
    return spam;
}


void spam_queue_free()
{
    Spam  *tmp, *sc;

    sc = spam;
    while (sc)  {
	tmp = sc;
	sc = sc->next;

	if (tmp->from_reg) regfree(tmp->from_reg);
	if (tmp->sub_reg)  regfree(tmp->sub_reg);
	
	free(tmp->from);
	free(tmp->subject);
	free(tmp);
    }
}


#ifdef	DEBUG

static void gmail_dump_att()
{
    int  i = 0;
    
    printf("Top lines:   %d\n", attr->topl);
    printf("Top style:   %d\n", attr->tops);
//    printf("quota:       %d\n", attr->quota);
    printf("login delay: %d\n", attr->delay);
    printf("screen cols: %d\n", attr->cols);
    printf("deft action: %d\n", attr->act);
    printf("set mbox:    %s\n", attr->mbox);
    printf("deft mbox:   %s\n", attr->defbox);
    printf("set user:    %s\n", attr->user);
    printf("set pass:    %s\n", attr->pass);
    printf("cp_set: ");
    for (i = 0; attr->cp_set && attr->cp_set[i]; i++)
	printf("%d,", attr->cp_set[i]);
    printf("\n");
    printf("headmask: ");
    for (i = 0; attr->hdmsk[i]; i++)
	printf("%s ", attr->hdmsk[i]);
    printf("\n\n");
}
	
static void gmail_dump(GMAIL *gm)
{
    Slnk  *sl;

    printf("GMAIL object %p\n", gm);
    printf("Remote host: %s\n", gm->host);
    printf("login id:    %s\n", gm->lid);
    printf("login pass:  %s\n", gm->pass);
    printf("set mbox:    %s\n", gm->mbox);
    printf("set action:  %d\n", gm->act);
    printf("protocol:    %d\n", gm->protl);
    printf("topspam:     %d\n", gm->topspam);
    for (sl = gm->slnk; sl; sl = sl->next)  {
	printf("spam id:     %d\n", sl->sc->id);
	printf("  quota:     %d\n", sl->sc->quota);
	printf("  from:      %s\n", sl->sc->from);
	printf("  subject:   %s\n", sl->sc->subject);
	printf("  logic:     %d\n", sl->sc->logic);
	printf("  action:    %d\n", sl->sc->action);
    }
    printf("\n");
}

int  spam_test(char *file)
{
    GMAIL *gm;
    Head  *head;
    FILE  *fp;
    char  buf[SVRBUF];
    Spam  *sp;

    gmail_dump_att();
    for (gm = gmail; gm; gm = gm->next)  gmail_dump(gm);
    
    file = line_refine(file, '#');
    fp = fopen(file, "r");
    if (fp == NULL)  return 1;

    head = head_register(HDPAGE, HDIDX);
    
    while (!feof(fp))  {
	
	head_clean(head);
	if (head_of_file(head, fp, buf, SVRBUF) == -1)  break;
	
	printf("-> %s\n", head->fmsp);
	for (gm = gmail; gm; gm = gm->next)  {
	    sp = spam_match_chain(gm->slnk, head, 0);
	    if (sp == NULL)  printf("test OK!\n");
	    else {
		printf("triggered! (%s)(%s)(%d)\n", 
				sp->from, sp->subject, sp->quota);
	    }
	}
    }	
    
    head_destruct(head);
    fclose(fp);
    return 0;
}

#endif    


