
/*
   codec.c -- implementation of base64 and quota-print encode/decode
   
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/********************** base64 encode/decode ****************************/

static const char base64[] = 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static const char index64[256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63, 
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1, 
	-1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1, 
	-1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 
};

static const char check64[256] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 
	0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 
	0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 
};

#define isbase64(c)	check64[(int) (c)]
#define char64(c)	index64[(int) (c)]

int  enbase64(char *sour, int n, char *dest)
{
    unsigned char  wb[3];
    int   cnt = 0;

    while (n >= 3)  {
	wb[0] = *sour++;
	wb[1] = *sour++;
	wb[2] = *sour++;
	
	*dest++ = base64[wb[0] >> 2];
	*dest++ = base64[((wb[0] & 0x3) << 4) | ((wb[1] & 0xf0) >> 4)];
	*dest++ = base64[((wb[1] & 0xf) << 2) | ((wb[2] & 0xc0) >> 6)];
	*dest++ = base64[wb[2] & 0x3f];

	n -= 3;
	cnt += 4;
    }
    if (n == 2)  {
	*dest++ = base64[sour[0] >> 2];
	*dest++ = base64[((sour[0] & 3) << 4) | ((sour[1] & 0xf0) >> 4)];
	*dest++ = base64[(sour[1] & 0xf) << 2];
	*dest++ = '=';
	cnt += 4;
    } else if (n == 1) {
	*dest++ = base64[sour[0] >> 2];
	*dest++ = base64[(sour[0] & 3) << 4];
	*dest++ = '=';
	cnt += 3;
    }
    return cnt;
}
    
int  debase64(char *sour, int n, char *dest)
{
    unsigned char  wb[4];
    int   cnt = 0;

    while (n >= 4)  {
	wb[0] = *sour++;
	wb[1] = *sour++;
	wb[2] = *sour++;
	wb[3] = *sour++;
	
	*dest++ = (wb[0] << 2) | (wb[1] >> 4);
	*dest++ = ((wb[1] & 0xf) << 4) | (wb[2] >> 2);
	*dest++ = ((wb[2] & 0x3) << 6) | wb[3];

	n -= 4;
	cnt += 3;
    }
    if (n == 3)  {
	*dest++ = (sour[0] << 2) | (sour[1] >> 4);
	*dest++ = ((sour[1] & 0xf) << 4) | (sour[2] >> 2);
	cnt += 2;
    } else if (n == 2)  {
	*dest++ = (sour[0] << 2) | (sour[1] >> 4);
	cnt++;
    }
    return cnt;
}
	
/* decode base64 string, ended by encountering '=' */

int  debase64_str(char *sour, char *buf)
{
    int  i = 0;
    char *p = buf;
    
    while (*sour && (*sour != '='))  {
	if (isbase64(*sour))  {
	    *p++ = char64(*sour);
	    i++;
	}
	sour++;
    }
    /* printf("%d> %s\n", i, buf); */
    return debase64(buf, i, buf);
}

/* Note that the 'glen' is three-octet-group, which means it reads file
   in three bytes a unit. The default is 18 and the encoded expands to 72 */

/*
int enbase64_file(FILE *fin, FILE *fout, int glen)
{
    char  *bin, *bout, *buf, cp1, cp2;
    int   len, off, k;

    if (glen == 0)  glen = 18;
    if ((buf = malloc(glen * 8)) == NULL)  return 1;
    bin = buf;
    bout = buf + glen * 3;

    off = 0, k = glen << 2;
    while ((len = fread(bin + off, 1, glen * 3 - off, fin)) > 0)  {
	off = len % 3;
	len = len / 3 * 3;
	cp1 = bin[len], cp2 = bin[len + 1];
		
	len = enbase64(bin, len, bout);
	if (len < k)  {
	    fwrite(bout, 1, len, fout);
	    k -= len;
	} else {
	    fwrite(bout, 1, k, fout);
	    fprintf(fout, "\n");
	    len -= k;
	    k = glen << 2;
	    if (len)  {
		fwrite(bout, 1, len, fout);
		k -= len;
	    }
	}

	bin[0] = cp1, bin[1] = cp2;
    }

    if (off)  {
	len = enbase64(bin, off, bout);
	if (len < k)  {
	    fwrite(bout, 1, len, fout);
	    k -= len;
	} else {
	    fwrite(bout, 1, k, fout);
	    fprintf(fout, "\n");
	    len -= k;
	    k = glen << 2;
	    if (len)  {
		fwrite(bout, 1, len, fout);
		k -= len;
	    }
	}
    }

    if (k == glen << 2)  fprintf(fout, "\n=\n");
    else fprintf(fout, "=\n");
    
    free(buf);
    return 0;
}
*/

int enbase64_file(FILE *fin, FILE *fout, int glen)
{
    char  bin[200], bout[256];
    int   i, ch = 0;

    if ((glen <= 0) || (glen > 60))  glen = 18;
    glen *= 3;
    while (ch != EOF)  {
	for (i = 0; i < glen; )  {
	    if ((ch = fgetc(fin)) == EOF)  break;
	    bin[i++] = ch;
	}
	i = enbase64(bin, i, bout);
	bout[i] = '\n';
	fwrite(bout, 1, i + 1, fout);
    }
    /* BUG: the '=' will appear at next line. no good, no bad */
    fprintf(fout, "=\n");
    return 0;
}

static char debuf[256];
static int  delen = 0;

static int  debase_fifo(FILE *fout, char *sour, int len)
{
    int  i, rs;

    if (sour == NULL)  {
	if (delen == 0)  return 0;
	rs = debase64(debuf, delen, debuf);
	fwrite(debuf, 1, rs, fout);
	delen = 0;
	return 0;
    } 
    
    for (i = 0; i < len; i++)  {
	if (sour[i] == '=')  {
	    rs = debase64(debuf, delen, debuf);
	    fwrite(debuf, 1, rs, fout);
	    delen = 0;
	    break;
	}
	if (delen == 256)  {
	    rs = debase64(debuf, delen, debuf);
	    fwrite(debuf, 1, rs, fout);
	    delen = 0;
	}
	if (isbase64(sour[i]))  {
	    debuf[delen++] = char64(sour[i]);
	}
    }
    
    return delen;
}


int debase64_file(FILE *fin, FILE *fout)
{
    char  buf[256];
    int   i, ch = 0;

    
    while ((ch != EOF) && (ch != '=')) {
	for (i = 0; i < 256; )  {
	    ch = fgetc(fin);
	    if ((ch == EOF) || (ch == '='))  break;
	    if (isbase64(ch))  buf[i++] = char64(ch);
	}
	i = debase64(buf, i, buf);
	fwrite(buf, 1, i, fout);
    }
    return 0;
}


char *debase64_bndry(FILE *fin, FILE *fout, char *bndry, char *buf, int len)
{
    int   blen;
    
    blen = strlen(bndry);
    
    while (fgets(buf, len, fin))  {
	if (!strncmp(buf, bndry, blen))  break;

	debase_fifo(fout, buf, strlen(buf));
    }
    debase_fifo(fout, NULL, 0);
    return buf;
}


/************************* Quoted-Printable ***********************/


#define QPLMT	76

static const char hexdec[256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1, -1, -1, -1, -1, -1, 
	-1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 
};

static const char dechex[17] = "0123456789ABCDEF";

#define hexnum(s)  ((int) hexdec[(int) (s)])
#define numhex(s)  dechex[s]
#define qphex(s)   ((hexnum(*(s)) << 4) | hexnum(*((s) + 1)))

int  atohex(char *s)
{
    int  rs;

    for (rs = 0; *s; s++)  {
	if (hexnum(*s) < 0)  break;
	rs = (rs << 4) | hexnum(*s);
    }
    return rs;
}

static const char qpfree[256] = {
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 1,  0,  0,  0,  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 
	 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  1,  1, 
	 0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 
	 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  1, 
	 0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 
	 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
};

#define isqpf(s)  qpfree[(int)(s)]

int enqp_str(unsigned char *sour, int n, char *buf)
{
    int  cur, len;

    if (n == 0)  return 0;
    cur = len = 0;
    while (n)  {
	if (isqpf(*sour))  {
	    if (len + 2 >= QPLMT)  {
		buf[cur++] = '=';
		buf[cur++] = '\n';
		len = 0;
	    }
	    buf[cur++] = *sour;
	    len++;
	} else if (*sour == '\n') {
	    if ((*(sour - 1) == ' ') || (*(sour - 1) == '\t'))  {
		buf[cur++] = '=';
		buf[cur++] = '\n';
	    }
	    buf[cur++] = *sour;
	    len = 0;
	} else {
	    if (len + 4 >= QPLMT)  {
		buf[cur++] = '=';
		buf[cur++] = '\n';
		len = 0;
	    }
	    buf[cur++] = '=';
	    buf[cur++] = numhex( *((unsigned char*) sour) >> 4);
	    buf[cur++] = numhex(*sour & 0xf);
	    len += 3;
	}
	n--;
	sour++;
    }
    if (*(sour - 1) != '\n')  {
	buf[cur++] = '=';
	buf[cur++] = '\n';
    }
    buf[cur] = 0;
    return cur;
}

int deqp_str(char *sour, char *buf)
{
    int  i, k;

    i = k = 0;
    while (sour[i]) {
	if (sour[i] == '=')  {
	    i++;
	    if (sour[i] == '\n')  i++;
	    else {
		buf[k++] = qphex(sour + i);
		i += 2;
	    }
	} else buf[k++] = sour[i++];
    }
    buf[k] = 0;		/* should be useless */
    return k;
}
	
int enqp_file(FILE *fin, FILE *fout)
{
    unsigned char bin[128], bout[512];
    int  i, ch = 0;

    while (ch != EOF)  {
	for (i = ch = 0; (i < 128) && (ch != '\n'); i++)  {
	    if ((ch = fgetc(fin)) == EOF)  break;
	    bin[i] = ch;
	}
	if (i)  {
	    i = enqp_str(bin, i, bout);
	    fwrite(bout, 1, i, fout);
	}
    }
    return 0;
}

int deqp_file(FILE *fin, FILE *fout)
{
    char bin[512], bout[512];
    int  rs;

    while (fgets(bin, 512, fin))  {
	rs = deqp_str(bin, bout);
	fwrite(bout, 1, rs, fout);
    }
    return 0;
}

char *deqp_bndry(FILE *fin, FILE *fout, char *bndry, char *buf, int len)
{
    char  bout[512];	/* hope it's enough */
    int   rs, blen;
    
    blen = strlen(bndry);
    
    while (fgets(buf, len, fin))  {
	if (!strncmp(buf, bndry, blen))  break;
	rs = deqp_str(buf, bout);
	fwrite(bout, 1, rs, fout);
    }
    return buf;
}

/************************** HZ <-> GB ***************************/

static const unsigned char gbhz[256] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 
	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 
	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 
	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 
	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 
	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 
	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 
	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 
	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 
	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0x7f, 
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 
	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 
	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0xff, 
};

int enhz_str(unsigned char *sour, int n, char *buf)
{
    int  cur, cf;

    cur = cf = 0;
    while (n) {
	if (*sour == '~')  {
	    buf[cur++] = '~';
	    buf[cur++] = '~';
	} else if ((*sour >= 0xa0) && (*sour <= 0xfe)) {
	    if (cf == 0) {
		cf = 1;
		buf[cur++] = '~';
		buf[cur++] = '{';
	    }
	    buf[cur++] = gbhz[*sour];
	} else {
	    if (cf == 1)  {
		cf = 0;
		buf[cur++] = '~';
		buf[cur++] = '}';
	    } 
	    buf[cur++] = *sour;
	}
	sour++;
	n--;
    }
    buf[cur] = 0;
    return cur;
}

int dehz_str(unsigned char *sour, char *buf)
{
    int  i, cf;

    i = cf = 0;
    while (*sour) {
	if (cf == 0)  {
	    if (*sour == '~')  {
		sour++;
		if (*sour == '~')  buf[i++] = '~';
		else if (*sour == '{')  cf = 1;
		else {
		    sour--;
		    buf[i++] = '~';
		}
	    } else buf[i++] = *sour;
	} else {
	    if (*sour == '~')  {
		sour++;
		if (*sour == '}')  cf = 0;
		else {
		    sour--;
		    buf[i++] = gbhz[*sour];
		}
	    } else buf[i++] = gbhz[*sour];
	}
	sour++;
    }
    buf[i] = 0;
    return i;
}
	
int enhz_file(FILE *fin, FILE *fout)
{
    unsigned char bin[256], bout[512];
    int  rs;

    while (fgets(bin, 256, fin))  {
	rs = enhz_str(bin, strlen(bin), bout);
	fwrite(bout, 1, rs, fout);
    }
    return 0;
}

int dehz_file(FILE *fin, FILE *fout)
{
    unsigned char bin[512], bout[512];
    int  rs;

    while (fgets(bin, 512, fin))  {
	rs = dehz_str(bin, bout);
	fwrite(bout, 1, rs, fout);
    }
    return 0;
}



#if	EXECUTABLE

#include "defs.h"

char ebcdic[] = "!\"#$@[\\]^`{|}~";
char litrl[] = "%&'()*+,-./0123456789:;<>?ABCDEFGHIJKLMNOPQRSTUVWXYZ_\
abcdefghijklmnopqrstuvwxyz \t";

static void enc_tab()
{
    int i, k, ch;
    char buf[256];

    printf("base64 character set\n\n");
    for (i = 0; i < 16; i++)  {
	printf("\t");
	for (k = 0; k < 16; k++)  {
	    ch = i * 16 + k;
	    if ((ch >= '0') && (ch <= '9')) ch = 1;
	    else if ((ch >= 'A') && (ch <= 'Z'))  ch = 1;
	    else if ((ch >= 'a') && (ch <= 'z'))  ch = 1;
	    else if ((ch == '+') || (ch == '/'))  ch = 1;
	    else ch = 0;
	    printf("%d, ", ch);
	}
	printf("\n");
    }

    memset(buf, -1, 256);
    for (i = 0; base64[i]; i++)  buf[(int) base64[i]] = i;
    printf("\nbase64 encode character set\n\n");
    for (i = 0; i < 16; i++)  {
	printf("\t");
	for (k = 0; k < 16; k++)  {
	    printf("%2d, ", buf[i*16+k]);
	}
	printf("\n");
    }
    
	    
    printf("\nGB <--> HZ codepage\n\n");
    for (i = 0; i < 32; i++)  {
	printf("\t");
	for (k = 0; k < 8; k++)  {
	    ch = i * 8 + k;
	    if ((ch >= 0x20) && (ch <= 0x7e))  ch += 0x80;
	    else if ((ch >= 0xa0) && (ch <= 0xfe))  ch -= 0x80;
	    printf("0x%02x, ", ch);
	}
	printf("\n");
    }

    printf("\nFor Quoted-Printable encodeing\n\n");
    for (i = 33; i < 127; i++)  {
	if (i == '=')  continue;
	for (k = 0; ebcdic[k]; k++)  
	    if (i == ebcdic[k])  break;
	if (ebcdic[k])  continue;
	putchar(i);
    }
    printf("\n");
    for (i = 0; i < 16; i++)  {
	printf("\t");
	for (k = 0; k < 16; k++)  {
	    int  j;
	    ch = i * 16 + k;
	    for (j = 0; litrl[j]; j++) 
		if (ch == litrl[j])  break;
	    if (litrl[j]) ch = 1;
	    else ch = 0;
	    printf("%2d, ", ch);
	}
	printf("\n");
    }
    
    printf("\nFor Hex-char to Dec-number\n\n");
    for (i = 0; i < 16; i++)  {
	printf("\t");
	for (k = 0; k < 16; k++)  {
	    ch = i * 16 + k;
	    if ((ch >= '0') && (ch <= '9'))       ch = ch - '0';
	    else if ((ch >= 'a') && (ch <= 'f'))  ch = ch - 'a' + 10;
	    else if ((ch >= 'A') && (ch <= 'F'))  ch = ch - 'A' + 10;
	    else ch = -1;
	    printf("%2d, ", ch);
	}
	printf("\n");
    }

/*    for (i = 0; i < 256; i++)  {
	if (qpfree[i])  putchar(i);
    }*/
}
 

static char *usage = "\
%s [-bBqQzZ] [--help] [--version] < infile > outfile\n\
\n\
   -b  encoding by base64\n\
   -B  decoding by base64\n\
   -q  encoding by quota-print\n\
   -Q  decoding by quota-print\n\
   -z  encoding GB file by HZ code\n\
   -Z  decoding HZ file to GB file\n\
   -t  coding table\n";
   
static char *ver = "%s 1.0\n";

int main(int argc, char **argv)
{
    int  op = 0;
    //char buf[256];
    
    if (argc < 2)  op = 'b';
    else if (!strcmp(argv[1], "--help"))   printf(usage, argv[0]);
    else if (!strcmp(argv[1], "--version"))  printf(ver, argv[0]);
    else if (!strcmp(argv[1], "-b"))  op = 'b';
    else if (!strcmp(argv[1], "-B"))  op = 'B';
    else if (!strcmp(argv[1], "-q"))  op = 'q';
    else if (!strcmp(argv[1], "-Q"))  op = 'Q';
    else if (!strcmp(argv[1], "-z"))  op = 'z';
    else if (!strcmp(argv[1], "-Z"))  op = 'Z';
    else if (!strcmp(argv[1], "-t"))  op = 't';
    else printf(usage, argv[0]);

    switch (op) {
    case 'b':
	enbase64_file(stdin, stdout, 0);
	break;

    case 'B':
	debase64_file(stdin, stdout);
	break;

    case 'q':
	enqp_file(stdin, stdout);
	break;

    case 'Q':
	deqp_file(stdin, stdout);
	break;

    case 'z':
	enhz_file(stdin, stdout);
	break;

    case 'Z':
	dehz_file(stdin, stdout);
	break;

    /*case 'h':
	op = 1;
	while (fgets(buf, 256, stdin))  {
	    if (!strncmp(buf, "set=", 4))  op = atoi(&buf[4]);
	    decode_head(buf, op);
	    puts(buf);
	}
	break;*/
	
    case 't':
	enc_tab();
	break;
    }
	
    return 0;
}

#endif


