
/*
   misc.c -- miscellany functions
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <ctype.h>
#include <netdb.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>


#define BLKBUF	4096

#define StrnCpy(d,s,l)  (strncpy((d),(s),(l)), (d)[(l)-1] = 0)

extern int ziptoken(char *sour, char **idx, int ids, char *delim);

static char blkbuf[BLKBUF];


#ifndef	HAVE_STRCASECMP

int strcasecmp(char *sour, char *dest)
{
    char    a, b;

    while (1)  {
	if ((*sour < 'a') || (*sour > 'z'))  
	    a = *sour;
	else  
	    a = *sour - 'a' + 'A';
	
	if ((*dest < 'a') || (*dest > 'z'))  
	    b = *dest;
	else  
	    b = *dest - 'a' + 'A';
	
	if (a != b)  
	    return 1;
	if (a == '\0')  
	    break;
	sour++, dest++;
    }
    return 0;
}


int strncasecmp(char *sour, char *dest, int leng)
{
    int     i;
    char    a, b;

    for (i = 0; i < leng; i++, sour++, dest++)  {
        if ((*sour < 'a') || (*sour > 'z'))  
	    a = *sour;
        else  
	    a = *sour - 'a' + 'A';
	
        if ((*dest < 'a') || (*dest > 'z'))  
	    b = *dest;
        else  
	    b = *dest - 'a' + 'A';
	
        if (a != b)  
	    return 1;
        if (a == '\0')  
	    break;
    }
    return 0;
}

#endif


char *dup_str(char *s)
{
    char *d, *tmp;

    if (!s)  
	return NULL;

    d = tmp = (char *) malloc(strlen(s) + 1);
    if ( d )
        while ((*d++ = *s++) != '\0');
    return tmp;
}

int dup_stream(FILE *fin, FILE *fout)
{
    int  rs, len = 0;

    while ((rs = fread(blkbuf, 1, BLKBUF, fin)) > 0)  {
        len += rs;
        fwrite(blkbuf, 1, rs, fout);
    }
    
    return len;
}


int isblankline(char *buf)
{
    while (*buf)  {
	if (isspace(*buf))  
	    buf++;
	else  
	    return 0;
    }
    return 1;
}


int isdigitword(char *buf)
{
    if (isblankline(buf))  
	return 0;
    
    while (*buf && isdigit(*buf))  
	buf++;
    
    if (!*buf || isspace(*buf))  
	return 1;
    return 0;
}

int isheadline(char *buf)
{
    if (isblankline(buf))  
	return 0;
    if (isspace(*buf))  
	return 1;
    if (strchr(buf, ':'))  
	return 1;
    return 0;
}

char *skip_space(char *sour)
{
    while (*sour && isspace(*sour))  
	sour++;
    return sour;
}


/* compress the inputing line, get rid of the comment and spare space
   at the beginning and ending parts. e.g:

   "   hello   # this is hello world " --> "hello"
*/

char *line_refine(char *s, int commt)
{
    char  *p;

    /* search the comment start */

    p = s;
    while ((p = strchr(p, commt)) != NULL)  {
        if ((p != s) && (*(p-1) == '\\'))  
	    continue;
        *p = '\0';
        break;
    }

    /* shorten the tail */

    p = s + strlen(s);
    while ((p != s) && (*p >= 0) && (*p <= ' '))  
	*p-- = '\0';

    /* compress the head */

    for (p = s; isspace(*p); p++);
    return p;
}


/* compress all white space characters to single space char. 
 * In the "C"  and "POSIX"   locales,   these  are:  space,  form-feed
 * ('\f'), newline  ('\n'),  carriage  return  ('\r'), horizontal tab 
 * ('\t'), and vertical tab ('\v'). */
        
void compress_space(char *sour)
{
    char *dest = sour;
    int  flag = 0;
    
    while (*sour)  {
	
	if (isspace(*sour))  {
	    if (flag == 0)  
		*dest++ = ' ';
	    flag++;
	} else {
	    flag = 0;
	    *dest++ = *sour;
	}
	sour++;
    }
    *dest = *sour;
}

/* get the last modification time of the specified file */

time_t filedate(char *filename, int *msize)
{
    struct stat stbuf;

    if (stat(filename, &stbuf) < 0)  return (time_t) -1;
    if (msize) *msize = stbuf.st_size;
    return stbuf.st_mtime;
}


static struct passwd  *loc_pwd  = NULL;
static struct hostent *loc_htt = NULL;
static char *loc_host = NULL;

char *get_user()
{
    if ( !loc_pwd )  loc_pwd = getpwuid(geteuid());
    return loc_pwd->pw_name;
}

char *get_home()
{
    if ( !loc_pwd )   loc_pwd = getpwuid(geteuid());
    return loc_pwd->pw_dir;
}

char *get_hostname()
{
    char buf[128];

    if ( !loc_host )  {
        gethostname(buf, sizeof(buf));
        if ((loc_htt = gethostbyname(buf)) == NULL)  {
            loc_host = malloc(strlen(buf) + 1);
            if ( loc_host )  strcpy(loc_host, buf);
        } else  loc_host = loc_htt->h_name;
    }
    return loc_host;
}


int ismailstart(char *buf)
{
    char        *idx[10], tmp[128];
    int         i, num;
    char        *week[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
    char        *mon[] = { "Jan", "Feb", "Mar",  "Apr", "May", "Jun",
                                "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

    StrnCpy(tmp, buf, 128);
    if ((num = ziptoken(tmp, idx, 10, " ")) < 7)  return 0;

    if (strcmp(idx[0], "From"))  return 0;

    for (i = 0; i < 7; i++)  if (!strcmp(idx[2], week[i]))  break;
    if (i == 7)  return 0;

    for (i = 0; i < 12; i++)  if (!strcmp(idx[3], mon[i]))  break;
    if (i == 12)  return 0;

    if (atoi(idx[4]) == 0)  return 0;
    if (strchr(idx[5], ':') == NULL)  return 0;
    if (atoi(idx[6]) == 0)  return 0;
    return 1;
}


int atonum(char *s)
{
    int	 tmp;

    tmp = atoi(s);
    while ((*s >= '0') && (*s <= '9'))  s++;
    switch (*s)  {
	case 'k':
	case 'K':
	    tmp *= 1024;
	    break;

	case 'm':
	case 'M':
	    tmp *= 1024 * 1024;
	    break;
    }
    return tmp;
}


/* store and adjust the mail download/handle set of which starts with 1. e.g.
   sour:       1,2,4-9,10,12         -->
   (int) set:  [1,1], [2,2], [4,9], [10,10], [12,12], [EOF,EOF]
*/

int isinset(int id, int *set)
{
    int  i;
    
    /* no SET means do not check, everything is going on */
    
    if (set == NULL)  return 1;
    
    for (i = 0; set[i] != EOF; i += 2)  {
    	if ((set[i] == 0) || (set[i+1] == 0))  continue;
    	if ((id >= set[i]) && (id <= set[i+1]))  return 1;
    	if ((id <= set[i]) && (id >= set[i+1]))  return 1;
    }
    return 0;
}


/* get current time like "Tue Jan 26 02:35:07 1999" */

static char ctm[32];

char *timenow()
{
    time_t  tm;

    time(&tm);
    strcpy(ctm, ctime(&tm));
    ctm [ strlen(ctm) - 1 ] = '\0';
    return ctm;
}

/* 
   if there is envirnment variety in the specified path, we should
   change it into absolute path. Note that '~' means home directory.
   i.e,  if home directory is /home/myhome:
   
      sour:                      dest:
   
      ~/Mail               ---> /home/myhome/Mail
      $HOME/Mail/.pmailrc  ---> /home/myhome/Mail/.pmailrc
*/	  

char *fullpath(char *sour, char *dest)
{
    char	buf[256], *idx[64], *env;
    int		i, num;

    StrnCpy(buf, sour, sizeof(buf));
    if (*buf == '/')  strcpy(dest, "/");  else  *dest = '\0';
    
    num = ziptoken(buf, idx, 64, "/");
    for (i = 0; i < num; i++)  {
    	if (!strcmp(idx[i], "~"))  {
    	    env = getenv("HOME");
    	    if (env)  strcat(dest, env);
    	} else if (idx[i][0] == '$')  {
    	    env = getenv(&idx[i][1]);
    	    if (env)  strcat(dest, env);
    	} else  strcat(dest, idx[i]);
	
	if (num - 1 != i)  strcat(dest, "/");
    }
    return dest;
}

