/*
   mkargs.c

   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#define ARGBUF	512

#define issym(x)	(((x) == '_') || isalnum(x))

static char *sour, *dest;
static int  dlen;		/* destination buffer limitation */


static void do_single_quota()
{
    sour++;	/* skip the leading single_quotation */

    while (*sour)  {
	if (dlen == 0)  break;

	switch (*sour)  {
	case '\'':
	    sour++;
	    return;
	    
	case '\\':
	    sour++;
	    if (*sour)  *dest++ = *sour++, dlen--;
	    break;

	default:
	    *dest++ = *sour++, dlen--;
	}
    }
}

static void do_env_var()
{
    int  i;
    char buf[128], *p;
    
    sour++;	/* skip the leading '$' sambol */

    if (*sour == '{')  {
	    
	sour++;
	for (i = 0; i < 128 && *sour && *sour != '}'; i++)   
	    buf[i] = *sour++;
	if (*sour == '}')  sour++;
	
    } else {
	    
	for (i = 0; i < 128 && issym(*sour); i++)   
	    buf[i] = *sour++;
	
    }
    buf[i] = '\0';
    
    if ((p = getenv(buf)) != NULL)  {
	strncpy(dest, p, dlen);
	dlen -= strlen(dest);
	dest += strlen(dest);
    }
}
	
static void do_double_quota()
{
    sour++;	/* skip the leading double quotation */

    while (*sour)  {
	if (dlen == 0)  break;
	
	switch (*sour)  {
	case '"':
	    sour++;
	    return;
	    
	case '\\':
	    sour++;
	    if (*sour)  *dest++ = *sour++, dlen--;
	    break;

	case '$':
	    do_env_var();
	    break;

	default:
	    *dest++ = *sour++, dlen--;
	}
    }
}

static void do_block()
{
    while (*sour)  {
	if (dlen == 0)  break;
	
	switch (*sour)  {
	case '"':
	    do_double_quota();
	    break;

	case '\'':
	    do_single_quota();
	    break;

	case '\\':
	    sour++;
	    if (*sour)  *dest++ = *sour++, dlen--;
	    break;
	    
	case '$':
	    do_env_var();
	    break;

	case '>':
	case '<':
	case '|':
	    goto b_end;
	    
	default:
	    if (*sour > ' ') *dest++ = *sour++, dlen--;
	    else goto b_end;
	}
    }

b_end:
    *dest = '\0';
}
    

int mkargs(char *sstr, char *argvs[], int argcs)
{
    char buf[ARGBUF];

    sour = sstr;
    while (*sour)  {
	    
	if ((*sour == '>') || (*sour == '<') || (*sour == '|'))  break;
	
	if (*sour > ' ')  {
		
	    /* init the globle variables */
    
	    dest = buf;
	    dlen = ARGBUF;
    
	    do_block();
	    
	    if ((*argvs = (char*) malloc(strlen(buf)+1)) == NULL)  return -1;
		
	    strcpy(*argvs, buf);
	    argvs++, argcs--;
	    
	    /* reserve a room for last argument, the NULL, of end */ 
	    if (argcs == 1)  break;
	    
	} else sour++;
    }
    *argvs = NULL;
    return argcs;
}


#if	EXECUTABLE

int main(int argc, char **argv)
{
    char *argvs[64], buf[256];
    int	 i, len;

    for (i = 0; i < argc; i++)  printf("arg [%2d]: %s\n", i, argv[i]);
    
    while (1)  {
	fgets(buf, 256, stdin);
	buf[strlen(buf) - 1] = 0;
	
	if (!strcmp(buf, "q"))  break;
	if (!strcmp(buf, "quit"))  break;
	if (!strcmp(buf, "exit"))  break;
	
	if ((i = mkargs(buf, argvs, 64)) < 0)  break;
	len = 64 - i;

	for (i = 0; i < len; i++)  {
	    printf("Arg [%2d]: %s\n", i, argvs[i]);
	    free(argvs[i]);
	}
    }
    return 0;
}
	
#endif

