
/* mktmpdir.c - create a unique temporary directory

   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* char *mktmpdir(char *template);
 * 
 * The  mktmpdir() function generates a unique temporary directory
 * name from template. The last six characters of template must be XXXXXX
 * and these are replaced with a string that makes the directory unique. 
 * The directory is then created with mode read/write and permissions 0755.
 * Mktmpdir() works quite like mkstemp() except it creates directory but file.
 * There is no race condition between testing and creating directory.
 */

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>


static const char letters[] =
"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

unsigned flip(unsigned u, int n)
{
    int i, ch;
    union {
	char      c[4];
	unsigned  u;
    } f;

    f.u = u;
    for (i = 0; i < n; i++)  {
	ch = f.c[0];
	f.c[0] = f.c[1];
	f.c[1] = f.c[2];
	f.c[2] = f.c[3];
	f.c[3] = ch;
	f.u += u;
    }
    return f.u;
}
    
unsigned shuffle(unsigned x, unsigned m)
{
    int  rs;

    rs = (int) x;
    if (rs < 0)
	rs = (rs << 1) ^ m;
    else
	rs = ((rs << 1) + 1) ^ m;
    return (unsigned) rs;
}

char *mktmpdir(char *tmpl)
{
    int  i, rs;
    char *p;
    unsigned int seed, pid;

    rs = strlen(tmpl);
    p = tmpl + rs - 6;
    if (rs < 6 || strcmp(p, "XXXXXX")) 
	return NULL;

    pid = flip(getpid(), 16);
    
    seed = rand() ^ pid;
    
    for (i = 0; i < 65535; i++)  {  	/* 65535 should be enough */
	p[0] = letters[seed % 62];
	seed = shuffle(seed, pid);
	p[1] = letters[seed % 62];
	seed = shuffle(seed, pid);
	p[2] = letters[seed % 62];
	seed = shuffle(seed, pid);
	p[3] = letters[seed % 62];
	seed = shuffle(seed, pid);
	p[4] = letters[seed % 62];
	seed = shuffle(seed, pid);
	p[5] = letters[seed % 62];
	seed = shuffle(seed, pid);
	
#ifdef	EXECUTABLE
	return tmpl;
#else
	if ((rs = mkdir(tmpl, 0755)) == 0)
	    return tmpl;
	else  if (errno != EEXIST)
	    return NULL;
#endif
	
    }

    return NULL;
}



#ifdef	EXECUTABLE

#include <stdio.h>

int main()
{
    int  i;
    char p[32];

    for (i = 0; i < 10; i++) {
	strcpy(p, "XXXXXX");
	mktmpdir(p);
	puts(p);
    }
    return 0;
}

#endif


    
