
/*
   progress.c -- display the rate of progress when download
   or upload something. 
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
   progree.c
   
   This module exports some procedure for indicating the progress
   of download/upload mail by displaying a string of dot in the 
   standard output (file description 1). This can be change in this
   file, which defined as OUTPUTOi.
   
   dotListInit()	every time to draw the progress, this
                    procedure should be called first
   dotIllustrate()  give it increased data length
   dotToEnd()		do not forget it when dot list is over

   Note that the last rate is average transfer rate. See dotToEnd()
   for details.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>

#if HAVE_UNISTD_H
  #include <sys/types.h>
  #include <unistd.h>
#endif

#if TIME_WITH_SYS_TIME
  #include <sys/time.h>
  #include <time.h>
#else
  #if HAVE_SYS_TIME_H
    #include <sys/time.h>
  #else
    #include <time.h>
  #endif
#endif


#define	OUTPUTTO	1


static int dotted = 0;	/* the number of printed dots */
static int dotall = 0;	/* the expected number for printing */

static int gotten = 0;	/* the byte number has fetched */
static int expect = 0;	/* the estimated length of file */

static struct timeval	start;	/* used for compute average tranfer */
                                /*  rate */


/* calculate the span of two time structure, return in milliseconds. */

static int diff_time(struct timeval *later, struct timeval *orig)
{
#if	defined(ARCH_i386) || defined(ARCH_i486)
    return (later->tv_sec - orig->tv_sec) << 10 + 
	    later->tv_usec >> 10 - orig->tv_usec >> 10;
#else
    return (later->tv_sec - orig->tv_sec) * 1000 + 
	    later->tv_usec / 1000 - orig->tv_usec / 1000;
#endif
}
    
    
/* produce data transfer rate, then store it in buf.
   the rate will be formated of constant length. i.e
   
      " 1024 bps" or " 10.2 K/s"
   Returned data stores in the buf, accurately 9 chars
*/
   
static void rate_format(int diff, int increased, char *buf)
{
    int	rate;

    /* if the data stream rate is too large or unexpectation happed,
       use violent as default infor */
    
    strcpy(buf, "  violent");
    if (diff <= 0)  
	return;
    
    rate = increased * 1000 / diff;
    
    /* append the denomination */
    
    if (rate < 10000)  {
	sprintf(buf, " %4d cps", rate);
    } else if (rate < 102400)  {
	sprintf(buf, " %d.%d K/s", rate >> 10, (rate & 0x3ff) >> 8);
    } else if (rate < 10240000)  {
	sprintf(buf, " %4d K/s", rate >> 10);
    } 
}

	
static void dot_prog_init(int exp, int dot_range)
{
    dotted  = 0;
    gotten  = 0;
    expect  = exp;
    dotall  = dot_range;
    gettimeofday(&start, (struct timezone *) NULL);
}

static void dot_prog_do(int d_in)
{
    int   dot_stop, i;
    char  rate[16];
    struct timeval last;
    
    if ( !expect || !dotall )  
	return;
    
    /* compute that how many dots should be printed by now */
    
    gotten += d_in;
    dot_stop = (dotall * gotten) / expect;
    if (dot_stop > dotall)  
	dot_stop = dotall; 
    
    if (dotted < dot_stop)  {
    	
    	/* print dots from last position to where it should be in */
    	
    	for (; dotted < dot_stop; dotted++)  
	    write(OUTPUTTO, ".", 1);
    	
	/* compute the time step increasing. */

	gettimeofday(&last, (struct timezone *) NULL);
	i = diff_time(&last, &start);
	rate_format(i, gotten, rate);
	
    	write(OUTPUTTO, rate, 9);
    	write(OUTPUTTO, "\b\b\b\b\b\b\b\b\b", 9);
    }
}

static void dot_prog_end()
{
    char rate[16];
    int  ms;
    struct timeval last;
    
    if ( !expect || !dotall )  
	return;
    
    for (; dotted < dotall; dotted++)  
	write(OUTPUTTO, ".", 1);
    
    gettimeofday(&last, (struct timezone *) NULL);
    ms = diff_time(&last, &start);
    rate_format(ms, gotten, rate);
    
    write(OUTPUTTO, rate, 9);
    write(OUTPUTTO, "\n", 1);
    
    dotted  = 0;
    gotten  = 0;
    expect  = 0;
    dotall  = 0;
}

static void pct_prog_do(int d_in)
{
    char rate[16];
    int  ms, i;
    struct timeval last;

    if (expect == 0)  
	return;

    gotten += d_in;
    i = sprintf(rate, "%3d%% ", gotten * 100 / expect);

    gettimeofday(&last, (struct timezone *) NULL);
    ms = diff_time(&last, &start);
    rate_format(ms, gotten, rate + i);

    write(OUTPUTTO, rate, 14);
    write(OUTPUTTO, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b", 14);
}

static void pct_prog_end()
{
    write(OUTPUTTO, "\n", 1);
    dotted  = 0;
    gotten  = 0;
    expect  = 0;
    dotall  = 0;
}

static void list_prog_do(int d_in)
{
    int  ms;
    struct timeval last;

    if (expect == 0)  
	return;

    gotten += d_in;
    gettimeofday(&last, (struct timezone *) NULL);
    ms = diff_time(&last, &start);

    if (ms)
	ms = gotten * 1000 / ms;
    printf("%d %d %d %d\n", d_in, gotten, expect, ms);
}
	
static void list_prog_end()
{
    dotted  = 0;
    gotten  = 0;
    expect  = 0;
    dotall  = 0;
}


void (*prog_init)(int exp, int range);
void (*prog_do)(int d_in);
void (*prog_end)();

void progress_init(int id)
{
    prog_init = dot_prog_init;
    
    switch (id) {
    case 1:			/* percentage display */
	prog_do = pct_prog_do;
	prog_end = pct_prog_end;
	break;
	
    case 2:
	prog_do = list_prog_do;
	prog_end = list_prog_end;
	break;
		
    default:
	prog_do = dot_prog_do;
	prog_end = dot_prog_end;
	break;
    }
}

#ifdef	DEBUG

void progress_test()
{
    int  i, k;
    
    for (k = 2; k >= 0; k--)  {
	progress_init(k);
	prog_init(100, 20);
	for (i = 0; i < 10; i++)  {
	    prog_do(10);
	}
	prog_end();
    }
}

#endif

