
/*
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

static int attr = '>';

static int min3(int a, int b, int c)
{
    if (a > b) 
	a = b;
    if (a > c)
	a = c;
    return a;
}

#define min(a,b)   (((a) < (b)) ? (a) : (b))
#define max(a,b)   (((a) < (b)) ? (b) : (a))

static char *skip_number(char *p)
{
    while (p) {
	switch (*p) {
	case '+':
	case '-':
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    p++;
	    break;
	default:
	    return p;
	}
    }
    return p;
}


/* this routine could be thought of putting a string inside a window, 
 * although the string might larger than the window. According to the 
 * polarity of 'exp', the string starts from the left side or ends at 
 * the right side of the window. So what we think about is the string's
 * effective length, the length of the padding space and the string's
 * displaying offset.
 *
 * Set window size is W, string length is S, buffer limit is N (=n-1).
 * Set effective string length is STRING, padding space is SPACE, offset
 * is OFFSET.
 *
 * Starts from left side: (exp < 0)
 *	STRING = min(W,S,N)
 *	SPACE  = max(0, min(W,N)-S)
 *	OFFSET = 0;
 * Ends at right side: (exp > 0)
 *      STRING = min(S,W) - MAX(0, W-N)
 *      SPACE  = max(0, min(W-S, N))
 *      OFFSET = max(0, S-W)
 * At center: (exp = 0)
 *      STRING = min(S,N)
 *      SPACE  = 0
 *      OFFSET = 0
 */

static int app_str(char *buf, int n, int exp, char *str)
{
    int  i, sln, len;
    char *p = NULL, *null = "(null)";

    n--;
    if (str == NULL)
	str = null;

    sln = strlen(str);
    
    if (exp < 0) {
	    
	exp = abs(exp);
	len = min3(exp, sln, n);
	for (i = 0; i < len; i++)
	    *buf++ = *str++;
	
	if (sln > len) {
	    p = buf - 1;
	    if (attr && len) 
		*p = attr;
	}
	
	len = max(0, min(exp, n) - sln);
	for (i = 0; i < len; i++)
	    *buf++ = ' ';
	
    } else if (exp == 0) {
	    
	len = min(sln, n);
	for (i = 0; i < len; i++)
	    *buf++ = *str++;
	
    } else {
	    
	len = max(0, min(exp - sln, n));
	for (i = 0; i < len; i++)
	    *buf++ = ' ';
	len = min(sln, exp) - max(0, exp - n);
	str += max(0, sln - exp);
	
	if (max(0, sln - exp))
	    p = buf;
	
	for (i = 0; i < len; i++)
	    *buf++ = *str++;

	if (attr && len && p)
	    *p = attr;
    }

    if (exp == 0)
	return min(n, sln);
    return min(n, exp);
}

	
static int app_percent(char *buf, int n, int exp)
{
    return app_str(buf, n, exp, "%");
}
	
static int app_char(char *buf, int n, int exp, int ch)
{
    char str[2];

    str[0] = ch;
    str[1] = 0;
    return app_str(buf, n, exp, str);
}

static int app_int(char *buf, int n, int exp, int ch)
{
    char str[32];

    sprintf(str, "%d", ch);
    return app_str(buf, n, exp, str);
}

static int app_unsigned(char *buf, int n, int exp, unsigned uch)
{
    char str[32];

    sprintf(str, "%u", uch);
    return app_str(buf, n, exp, str);
}

static int app_float(char *buf, int n, int exp, double f)
{
    char str[64];

    sprintf(str, "%f", f);
    return app_str(buf, n, exp, str);
}

static int app_point(char *buf, int n, int exp, void *p)
{
    char str[32];

    sprintf(str, "%p", p);
    return app_str(buf, n, exp, str);
}

static int app_hex(char *buf, int n, int exp, int ch, int hex)
{
    char str[32];
    int  i, tmp;

    str[31] = 0;
    for (i = 30; i >= 0; i--)  {
	tmp = ch & 0xf;
	if (tmp < 10)
	    str[i] = tmp + '0';
	else if (hex == 'x')
	    str[i] = tmp - 10 + 'a';
	else 
	    str[i] = tmp - 10 + 'A';
	ch = ch >> 4;
    }
    return app_str(buf, n, exp, &str[i]);
}
	
/* %-9d %s %f %% %c %u %x %X %p */
/* ssprintf(buf, BUFSIZ, "%9d %-13s %10p\n", i, s, p); */

int ssprintf(char *buf, int n, char *fmt, ...)
{
    va_list ap;
    int  i, rs, len, ss;
    char *p;
    unsigned u;
    float f;
    
    va_start(ap, fmt);

    ss = len = 0;
    while (*fmt)  {
	if (n == 1)	/* the last ASC0 must be reserved */
	    break;
	
	if (ss == 0) {
		
	    if (*fmt == '%') {
		ss = 1;
		fmt++;
	    } else {
		*buf++ = *fmt++;
		len++;
		n--;
	    }
	    
	} else {
		
	    rs = atoi(fmt);
	    fmt = skip_number(fmt);
	    switch (*fmt) {
	    case 0:
		break;
		
	    case '%':
		rs = app_percent(buf, n, rs);
		break;
		
	    case 'c':
		i = va_arg(ap, int);
		rs = app_char(buf, n, rs, i);
		break;
		
	    case 'd':
	    case 'l':
		i = va_arg(ap, int);
		rs = app_int(buf, n, rs, i);
		break;
		
	    case 'x':
	    case 'X':
		i = va_arg(ap, int);
		rs = app_hex(buf, n, rs, i, *fmt);
		break;
		
	    case 's':
		p = va_arg(ap, char*);
		rs = app_str(buf, n, rs, p);
		break;
		
	    case 'p':
		p = va_arg(ap, void*);
		rs = app_point(buf, n, rs, p);
		break;
		
	    case 'u':
		u = va_arg(ap, unsigned);
		rs = app_unsigned(buf, n, rs, u);
		break;
		
	    case 'f':
		f = va_arg(ap, double);
		rs = app_float(buf, n, rs, f);
		break;
		
	    default:
		rs = 0;
		break;
	    }
	    
	    if (*fmt)  {
		buf += rs;
		n -= rs;
		fmt++;
	    }
	    ss = 0;
	}
    }
    va_end(ap);
    *buf = 0;
    return len;
}


/* %d %s %f %% %c %u %x %X %p */
/* siprintf(buf, BUFSIZ, "%d %s %p\n", 9, i, -13, s, 10, p); */

int siprintf(char *buf, int n, char *fmt, ...)
{
    va_list ap;
    int  i, rs, len, ss;
    char *p;
    unsigned u;
    float f;
    
    va_start(ap, fmt);

    ss = len = 0;
    while (*fmt)  {
	if (n == 1)	/* the last ASC0 must be reserved */
	    break;
	
	if (ss == 0) {
		
	    if (*fmt == '%') {
		ss = 1;
		fmt++;
	    } else {
		*buf++ = *fmt++;
		len++;
		n--;
	    }
	    
	} else {
		
	    fmt = skip_number(fmt);
	    rs = va_arg(ap, int);
	    switch (*fmt) {
	    case 0:
		break;
		
	    case '%':
		rs = app_percent(buf, n, rs);
		break;
		
	    case 'c':
		i = va_arg(ap, int);
		rs = app_char(buf, n, rs, i);
		break;
		
	    case 'd':
	    case 'l':
		i = va_arg(ap, int);
		rs = app_int(buf, n, rs, i);
		break;
		
	    case 'x':
	    case 'X':
		i = va_arg(ap, int);
		rs = app_hex(buf, n, rs, i, *fmt);
		break;
		
	    case 's':
		p = va_arg(ap, char*);
		rs = app_str(buf, n, rs, p);
		break;
		
	    case 'p':
		p = va_arg(ap, void*);
		rs = app_point(buf, n, rs, p);
		break;
		
	    case 'u':
		u = va_arg(ap, unsigned);
		rs = app_unsigned(buf, n, rs, u);
		break;
		
	    case 'f':
		f = va_arg(ap, double);
		rs = app_float(buf, n, rs, f);
		break;
		
	    default:
		rs = 0;
		break;
	    }
	    
	    if (*fmt)  {
		buf += rs;
		n -= rs;
		fmt++;
	    }
	    ss = 0;
	}
    }
    va_end(ap);
    *buf = 0;
    return len;
}


void ssp_def(int ch)
{
    attr = ch;
}

#ifdef	EXECUTABLE

int main()
{
    char buf[128];

    ssprintf(buf,10,"%s\n", "asdfgh");
    printf(buf);
    ssprintf(buf,10,"%10s\n", "asdfgh");
    printf(buf);
    ssprintf(buf,10,"%10s\n", "asdfghijklmn");
    printf(buf);
    siprintf(buf, 128, "%d %s %p\n", 9, 12, -13, "asdfasdf", 10, buf);
    printf(buf);
    return 0;
}

#endif

