
/*
   main.c 
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include "defs.h"
#include "gmail.h"


/* these variables are set against command line arguments */

char  *cmd_auth = NULL;		/* authorization pair */
char  *cmd_host = NULL;
char  *cmd_mbox = NULL;
int   *cmd_cp_set = NULL;	/* download set */
int   cmd_action = 0;

static char  *cmd_subject = NULL;
static char  *cmd_cc = NULL;
static char  *cmd_bcc = NULL;		/* blind carbon copy */

/* these variables are set against option file */

int	timeout = 90;
int	verbose = 0;

static int defbeh = ACT_DEFT;	/* index use default behave */


static int  parse_command_line(int argc, char **argv, char *abuf, int alen);
static int  *set_download_set(char *s);
static int  set_default(char *s, int *tasks, int tlen);
static int  open_quota_log(char *s);
static void signal_quit(int sig);
static void append_template();
static void usage(int flag);


static char *rcpath[] = {
	NULL,
	"/etc/pmailrc",
	"/usr/lib/pmailrc",
	"/usr/local/etc/pmailrc",
	"/usr/local/lib/pmailrc",
	NULL
};

int main(int argc, char **argv)
{
    Folder *fo = NULL;
    SMTP   *sm;
    GMAIL  *gm;
    Head   *head;
    int	   i, rs, tasks[TASK_NR];
    char   *abuf, *p;
    
    umask(077);
    
    signal(SIGHUP, signal_quit);
    signal(SIGINT, signal_quit);
    signal(SIGQUIT, signal_quit);
    signal(SIGTERM, signal_quit);
    signal(SIGPIPE, signal_quit);
    signal(SIGALRM, signal_quit);
    
    //progress_test();
    progress_init(0);
    
    /* create a buffer for gathering command line arguments */
    
    abuf = malloc(BLKBUF);
    if (abuf == NULL)  
	return -1;
    
    /* get the home directory, then open default RC file. */
    
    sprintf(abuf, "%s/.pmailrc", get_home());
    rcpath[0] = abuf;

    /* if DEBUG is on, Pmail should search in the current directory first
     * for the configuration file. It makes debugging life easier since we can
     * test options in working directory.
     * A distribution version should disable it. */
    
    for (i = 0; rcpath[i] != NULL; i++)  {
	if (rc_open(rcpath[i]) >= 0)  
	    break;
    }
    
    if ((p = rc_check(RCKEY_TIMEOUT)) != NULL)  
	timeout = atoi(p);
    if ((p = rc_check(RCKEY_VERBOSE)) != NULL)  
	verbose = 1;
    if ((p = rc_check(RCKEY_SMBEHV)) != NULL)
	defbeh  = msg_behave_bind(p);
    if ((p = rc_check(RCKEY_DEBUG)) != NULL)
	open_quota_log(p);
    
    /* now parsing command line arguments.
     * It is supposed that you can list a series of tasks in the config file, 
     * in the 'default' option. You can also overwhelm this task queue by an
     * alternative instruction in the command line. In that case, the task
     * queue is fixed to one element, the instruction. */
    
    rs = parse_command_line(argc, argv, abuf, BLKBUF);
    if (rs != ID_NULL)  {
	tasks[0] = rs, rs = 1;
    } else if ((p = rc_check(RCKEY_DEFAULT)) != NULL)  {
	rs = set_default(p, tasks, TASK_NR);
    } else rs = 0;
    
    /* process the task queue */
    
    for (i = 0; i < rs; i++)  {
	/* printf("Task %d/%d: %d\n", i, rs, tasks[i]); */
	    
	switch(tasks[i])  {
	case ID_HELP:
	    usage(ID_HELP);
	    break;

	case ID_VERSION:
	    usage(ID_VERSION);
	    break;

	case ID_TEMPLATE:
	    append_template();
	    break;

	case ID_GMAIL:
	    //progress_init(0);
	    head = head_register(HDPAGE, HDIDX);
	    if (head == NULL)
		break;
	    gm = gmail_open();
	    
	    if (cmd_host)  {
		gm = gmail_cmdarg(cmd_host);
		if (gm)  
		    gmail_single(gm, head);
	    } else  
		gmail_main(gm, head);

	    gmail_close();
	    head_destruct(head);
	    break;

	case ID_SMTP:
	    if (*abuf)  {
		fo = folder_open(abuf, defbeh);
	    }
	    if (cmd_mbox)  {
		fo = folder_open(cmd_mbox, defbeh);
	    }
	    if (fo == NULL)  {
		p = rc_check(RCKEY_OUTBOX);
		if (p != NULL)  {
		    fo = folder_open(p, defbeh);
		}
	    }
		    
	    rs = folder_number();
	    if (rs < 0)  {
		conPrintf("Warning: No postpone message folder specified.\n");
	    } else if (rs == 0)  {
		conPrintf("No message is ready for sending.\n");
	    } else {
		conPrintf("%d %s ready for sending.\n", 
			    rs, rs < 2 ? "message is" : "messages are");
	
		sm = smtp_host_chain(cmd_host, cmd_auth);
		smtp_main(fo, sm);
		smtp_host_clean();
	    }
	    
	    folder_close(rc_check(RCKEY_BACKUP), defbeh);
	    break;

	case ID_STREAM:
	    fo = folder_online(abuf, cmd_cc, cmd_bcc, cmd_subject, defbeh);
	    if (fo == NULL)  
		break;
	    sm = smtp_host_chain(cmd_host, cmd_auth);
	    smtp_main(fo, sm);
	    smtp_host_clean();
	    folder_close_online(rc_check(RCKEY_BACKUP), defbeh);
	    break;

	case ID_HOSTLIST:
	    gmail_open();
	    gmail_host_list();
	    gmail_close();
	    
	    sm = smtp_host_chain(cmd_host, cmd_auth);
	    smtp_host_list();
	    smtp_host_clean();
	    break;

	case ID_DUMPTMP:
	    if (*abuf)  {
		fo = folder_open(abuf, defbeh);
	    } 
	    if (cmd_mbox)  {
		fo = folder_open(cmd_mbox, defbeh);
	    }
	    if (fo == NULL)  {
		p = rc_check(RCKEY_OUTBOX);
		if (p != NULL)  {
		    fo = folder_open(p, defbeh);
		}
	    }
		    
	    folder_dump(fo);
	    /* getchar(); */

	    folder_close(NULL, ACT_HOLD);
	    break;

	case ID_DUMPINFO:
	    if (*abuf)  {
		fo = folder_open(abuf, defbeh);
	    }
	    if (cmd_mbox)  {
		fo = folder_open(cmd_mbox, defbeh);
	    }
	    if (fo == NULL)  {
		p = rc_check(RCKEY_OUTBOX);
		if (p != NULL)  {
		    fo = folder_open(p, defbeh);
		}
	    }
		    
	    folder_list(fo);

	    folder_close(NULL, ACT_HOLD);
	    break;

#ifdef	DEBUG
	case ID_SPAMTEST:
	    gm = gmail_open();
	    if (cmd_host)  
		gm = gmail_cmdarg(cmd_host);

	    spam_test(abuf);
	    gmail_close();
	    break;
#endif
	    
	}
    }
    
    free(abuf);
    ReleaseResource();
    return 0;
}


static int parse_command_line(int argc, char **argv, char *abuf, int alen)
{
    int   oidx, c, rs;
    char  *sopt="pPvotf:h:d:s:c:b:l:e:u:";
    struct option lopt[] = {
    	{ "verbose",	0, NULL, 'v' },
    	{ "behave",	1, NULL, 'e' },
    	{ "host",	1, NULL, 'h' },
    	{ "download",	1, NULL, 'd' },
    	{ "list",	1, NULL, 'l' },
	{ "auth",	1, NULL, 'u' },
    	{ "template",	0, NULL, 't' },
    	{ "help",	0, NULL, ID_HELP },
    	{ "version",	0, NULL, ID_VERSION },
	{ "spamtest",	0, NULL, ID_SPAMTEST },
	{ "server",	0, NULL, ID_HOSTLIST },
	{ "progress",	1, NULL, ID_PROGRESS },
    	{ 0, 0, 0, 0 }
    };

    rs = ID_NULL;
    while (1)  {
    	c = getopt_long(argc, argv, sopt, lopt, &oidx);
    	if (c == -1)  break;
    	switch (c)  {
    	    case ID_HELP:
    		return ID_HELP;
    		
    	    case ID_VERSION:
    		return ID_VERSION;

	    case ID_HOSTLIST:
		rs = ID_HOSTLIST;
		break;
		
#ifdef	DEBUG
	    case ID_SPAMTEST:
		rs = ID_SPAMTEST;
		break;
#endif
		
	    case ID_PROGRESS:
		progress_init(atoi(optarg));
		break;
    		
    	    case 'v':
    		verbose = YES;
    		break;

	    case 'a':
		defbeh = msg_behave_bind(optarg);
		break;
    		
    	    case 't':
    		rs = ID_TEMPLATE;
		break;
    		
	    case 'P':
		rs = ID_DUMPTMP;
		break;

	    case 'p':
		rs = ID_DUMPINFO;
		break;

    	    case 'o':
    		rs = ID_SMTP;
    		break;
    		
    	    case 'h':
		if (cmd_host == NULL)  cmd_host = dup_str(optarg);
    		break;
    		
    	    case 'f':
		if (cmd_mbox == NULL)  cmd_mbox = dup_str(optarg);
    		break;
    		
    	    case 's':
		if (cmd_subject == NULL)  cmd_subject = dup_str(optarg);
    		break;
    		
    	    case 'c':
		if (cmd_cc == NULL)  cmd_cc = dup_str(optarg);
    		break;
    		
    	    case 'b':
		if (cmd_bcc == NULL)  cmd_bcc = dup_str(optarg);
    		break;
    					
	    case 'u':
		if (cmd_auth == NULL)  cmd_auth = dup_str(optarg);
		break;

    	    case 'd':
		rs = ID_GMAIL;
		cmd_action = gmail_set_action(optarg);
		break;
    		
    	    case 'l':
    		cmd_cp_set = set_download_set(optarg);
    		break;
    		
    	    default:
    		return ID_HELP;
    	}
    }	
 
    *abuf = '\0';
    if (optind < argc)  {
	for (c = optind; c < argc; c++)  {
	    if (strlen(abuf) + strlen(argv[c]) + 1 < alen)  {
		strcat(abuf, argv[c]);
		strcat(abuf, " ");
	    }
	}
	if (rs == ID_NULL)  rs = ID_STREAM;
    }
    return rs;    
}


/* store and adjust the mail download/handle set of which starts with 1. e.g.
   sour:          1,2,4-9,10,12         -->
   download set:  [1,1], [2,2], [4,9], [10,10], [12,12], [-1,-1]    (int array)
*/
static int *set_download_set(char *s)
{
    char	*idx[PARMAX], *p;
    int 	*dl, max, i;
    
    /* first it acknowleges the number of ',' */
    
    for (max = 0, p = s; (p = strchr(p, ',')) != NULL;  p++, max++);
    
    max = ((max + 2) * sizeof(int)) << 1;
    if ((dl = malloc(max)) == NULL)  return NULL;
    
    max = ziptoken(s, idx, PARMAX, ",");
    for (i = 0; i < max; i++)  {
    	dl[i*2] = atoi(idx[i]);
    	if ((p = strchr(idx[i], '-')) == NULL)  dl[i*2+1] = dl[i*2];
	else dl[i*2+1] = atoi(p+1);
    }
    dl[i*2] = dl[i*2+1] = EOF;		/* -1, end flag */
    return dl;
}

static int set_default(char *s, int *tasks, int tlen)
{
    char	*idx[TASK_NR];
    int		i, num;
    
    /* this function read a option, 'default xxx,xxx,...', from 
       RC file. If there were another action specified by command
       line arguments, the 'default' option should be ignored. */

    num = ziptoken(s, idx, TASK_NR, DELIMS);
    
    for (i = 0; (i < num) && (i < tlen); i++)  {

    	if (!strcasecmp(idx[i], "get"))           tasks[i] = ID_GMAIL;
    	else if (!strcasecmp(idx[i], "sent"))     tasks[i] = ID_SMTP;
    	else if (!strcasecmp(idx[i], "template")) tasks[i] = ID_TEMPLATE;
	else if (!strcasecmp(idx[i], "dump"))     tasks[i] = ID_DUMPTMP;
	else if (!strcasecmp(idx[i], "info"))     tasks[i] = ID_DUMPINFO;
    	else if (!strcasecmp(idx[i], "version"))  tasks[i] = ID_VERSION;
    	else if (!strcasecmp(idx[i], "help"))     tasks[i] = ID_HELP;
	else i--;
	    
	/* printf("default: %s = %d\n", idx[i], tasks[i]); */
    }
    return i;
}

static int  open_quota_log(char *s)
{
    char *idx[2];
    char buf[SVRBUF];
    int  rs, quota;

    quota = 0;
    rs = ziptoken(s, idx, 2, DELIMS);
    if (idx[1] && *idx[1])  {
        quota = atoi(idx[1]);
        switch (idx[1][strlen(idx[1]) - 1])  {
        case 'k':
        case 'K':
            quota *= 1024;
            break;

        case 'm':
        case 'M':
            quota *= 1024 * 1024;
            break;
        }
    }
    fullpath(idx[0], buf);
    //printf("open log: %s %d\n", buf, quota);
    return log_open(buf, quota);
}


static void signal_quit(int sig)
{
    ReleaseResource();
    switch(sig)  {
    	case SIGHUP:
    	    printf("\nEncountered signal SIGHUP (%d)\n", sig);
    	    break;
    		
    	case SIGINT:
    	    printf("\nEncountered signal SIGINT (%d)\n", sig);
    	    break;
    		
    	case SIGQUIT:
    	    printf("\nEncountered signal SIGQUIT (%d)\n", sig);
    	    break;
    		
    	case SIGTERM:
    	    printf("\nEncountered signal SIGTERM (%d)\n", sig);
    	    break;
    		
    	case SIGALRM:
    	    printf("\nEncountered timeout signal (%d)\n", sig);
    	    break;
    		
    	case SIGPIPE:
    	    printf("\nEncountered signal SIGPIPE (%d)\n", sig);
    	    break;
    }
    exit(sig);
}


static void append_template()
{
    FILE  *fp = NULL;
    char  obox[SVRBUF], *tmp;
    
    if ((tmp = cmd_mbox) == NULL)  {
	tmp = rc_check(RCKEY_OUTBOX);
    }
    if (tmp != NULL)  {
	fullpath(tmp, obox);
	fp = fopen(obox, "a");
    }
    if (fp == NULL)  fp = stdout;

    tmp = rc_check(RCKEY_REPLY);
    if (tmp == NULL)  {
    	fprintf(fp, "\nFrom - %s\n", timenow());
    	fprintf(fp, "From:\n");
    } else {
    	fprintf(fp, "\nFrom %s %s\n", email_pure(tmp), timenow());
    	fprintf(fp, "From: %s\n", tmp);
    }
    fprintf(fp, "To: \n");
    fprintf(fp, "X-behave: pipe,sign,exhd\n");
    fprintf(fp, "Subject: \n\n");
    if (fp != stdout)  fclose(fp);
    printf("Simple template is appended into %s.\n",
    		obox ? obox : "stdout");
}
    


/**************************************************************/
    	
static void usage(int flag)
{
    char *ver="pmail %s, by Xuming <xuming@iname.com>\n";

    char *msg = "\
Usage:\n\
 pmail [-v] [-e smtp-beh] [-u auth-pair] [-f mailbox] [-h host] [-l ...]\n\
\n\
 pmail -dd [...]  download mails and remove them in the server\n\
 pmail -df [...]  fetch mails and reserve them in the server\n\
 pmail -dc [...]  check mailbox status in the server\n\
 pmail -dl [...]  list message heads or other informations\n\
 pmail -dr [...]  remove mails in the server end\n\
                  You can instruct more than one POP3 server in [...]\n\
\n\
 pmail -o [...]   send mails stored in the postpone folder\n\
 pmail -p [...]   preview the messages' head in the postpone folder\n\
 pmail -P [...]   dump the postpone folder to screen\n\
                  You can instruct more than one postpone folder in [...]\n\
\n\
 pmail [-s subject] [-c cc-addr] [-b bcc-addr] to-addr [...]\n\
                  You can instruct more than one receiver in [...]\n\
\n\
 pmail -t         appends a template to the postpone mailbox (obsolete)\n\
\n\
";
   
    if (flag == ID_VERSION)  printf(ver, VERSION); else  printf(msg);
}
    
/**********************************************************/


/* this procedure used to release all of the source which pmail ever
   occured. It also is looked on the last defend line as signal 
   encountered or other urgent affair occured */

void ReleaseResource()
{
    log_close();
    rc_close();
    folder_close(NULL, ACT_HOLD);
   // clear_gmail();
    
    xfree(cmd_host);
    xfree(cmd_mbox);
    xfree(cmd_subject);
    xfree(cmd_cc);
    xfree(cmd_bcc);
    xfree(cmd_cp_set);
}

