
/*
   memo.c  -- handle the address memo. 
   
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "defs.h"
#include "memo.h"


Memo *memo_open(char *memo)
{
    Memo  *mm;
    char  buf[SVRBUF];

#ifdef	DEBUG
    if (!memo)  return NULL;
#endif

    if ((mm = malloc(sizeof(Memo))) == NULL)  {
	dbgPrintf("memo_open: out of memory.\n");
	return NULL;
    }
    mm->num = 0;
    
    if ((mm->fp = fopen(memo, "a+")) == NULL)  {
	dbgPrintf("memo_open: Cannot open address memo. [%s]\n", memo);
	free(mm);
	return NULL;
    }

    if ((mm->pm = pm_handler(PGIDX, PGSIZE)) == NULL)  {
	dbgPrintf("memo_open: No PMAL handle allocated.\n");
	fclose(mm->fp);
	free(mm);
	return NULL;
    }

    rewind(mm->fp);
    while (fgets(buf, SVRBUF, mm->fp))  {
	line_refine(buf, '#');
	if (isblankline(buf))  continue;

	if (mm->num >= MMIDX)  break;
	mm->idx[mm->num] = pm_dup_str(mm->pm, buf);
	if (mm->idx[mm->num])  mm->num++;
    }
	    
    return mm;
}

void memo_close(Memo *mm)
{
#ifdef	DEBUG
    if (!mm)  return;
#endif
    
    fclose(mm->fp);
    pm_close(mm->pm);
    free(mm);
}
	

int memo_add_addr(Memo *mm, char *addr)
{
    int   i;

#ifdef	DEBUG
    if (!mm || !addr)  return 1;
#endif
    
    if (isblankline(addr))  return 0;

    for (i = 0; i < mm->num; i++)  {
	if (!email_cmp(mm->idx[i], addr))  return 0;  /* existed already */
    }
    
    if (mm->num >= MMIDX)  return 0;
    mm->idx[mm->num] = pm_dup_str(mm->pm, addr);
    if (mm->idx[mm->num])  mm->num++;
	    
    fseek(mm->fp, 0, SEEK_END);
    fprintf(mm->fp, "%s\n", addr);
    fflush(mm->fp);

    return 1;
}


char *memo_choose(Memo *mm, int col, int line)
{
    char  buf[SVRBUF];
    int   i, k, len;
    
#ifdef	DEBUG
    if (!mm)  return NULL;
#endif

    i = 0;
    line--;
    while(1) {
	for (k = 0; (i < mm->num) && (k < line); i++, k++)  {
	    len = sprintf(buf, "%4d- ", i + 1);
	    StrnCpy(buf + len, mm->idx[i], SVRBUF - len);
	    len = strlen(buf);
	    if (len > col)  {
		buf[col - 2] = '-';
		buf[col - 1] = '\0';
	    }
	    printf("%s\n", buf);
	}
	
	printf("'b' back, 'f' forward, 'q' quit, 'a' abort, choose: ");
	fgets(buf, SVRBUF, stdin);
	line_refine(buf, '#');
	
	if (!strcasecmp(buf, "q"))  return NULL;
	else if (!strcasecmp(buf, "a"))  return (char*) memo_choose;
	else if (!strcasecmp(buf, "b")) {
	    i -= line;
	    if (i < 0)  i = 0;
	} else if (isdigit(buf[0]))  {
	    i = atoi(buf) - 1;
	    if (i < 0)  return NULL;
	    else  return mm->idx[i];
	} else if (i == mm->num)  {
	    i -= line;
	    if (i < 0)  i = 0;
	}
    }
    return NULL;
}
    
    
