
/*
   mhead.c -- Internet message head general processing routine.
	
   Copyright (C) 1998, 2000  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"

static char *chop_tail(char *p);
	

Head *head_register(int page, int idx)
{
    Head  *tmp;

    if ((tmp = malloc(sizeof(Head))) == NULL)  return NULL;
    memset(tmp, 0, sizeof(Head));

    tmp->pm = pm_handler(idx, page);
    if (tmp->pm == NULL)  {
	free(tmp);
	return NULL;
    }

    tmp->this = tmp;
    return tmp;
}

void head_destruct(Head *hdo)
{
    if (hdo != NULL)  {		/* all memory allocating routines are issued*/
	pm_close(hdo->pm);	/* by pmalloc.c, therefore no more releasing*/
	free(hdo->this);	/* procession required */
    }
}

char *head_loading(Head *hdo, char *s)
{
    struct head_chain  *hc, *hctmp;
    HLin   *hltmp;
    int    len;
    
    if (!isheadline(s))  return s;

    /* continuing last head line
       If a headline starts with a white space character such as SPACE,
       H-TAB, it should be the remain part of last head line */
    
    if (isspace(*s))  {	

	/* it should not be the first of all the head part, otherwise
	   it declares the end of head part */
	    
	if (hdo->last == NULL)  return s;
	
	hctmp = pm_alloc(hdo->pm, sizeof(struct head_chain));
	if (hctmp == NULL)  return s;

	hctmp->next = NULL;
	hctmp->str = pm_dup_str(hdo->pm, s);
	
	if (hdo->last->hc == NULL)  hdo->last->hc = hctmp;
	else {
	    for (hc = hdo->last->hc; hc->next; hc = hc->next);
	    hc->next = hctmp;
	}
	
	if (hctmp->str)  {
	    len = strlen(hctmp->str);
	    hdo->last->len += len;
	    hdo->len += len;
	}
	return NULL;
    }

    /* create new head line structure */
    
    hltmp = pm_alloc(hdo->pm, sizeof(HLin));
    if (hltmp == NULL)  return s;
    
    hltmp->rm = 0;
    hltmp->len = len = strlen(s);
    hltmp->hc = hltmp->next = NULL;
    hltmp->headin = pm_dup_str(hdo->pm, s);
    if (hltmp->headin == NULL)  {
	hltmp->body = NULL;
	hltmp->key = NULL;
    } else {
	hltmp->body = strchr(hltmp->headin, ':');
	*hltmp->body = 0;
	hltmp->key = pm_dup_str(hdo->pm, hltmp->headin);
	chop_tail(hltmp->key);
	*hltmp->body = ':';
	for (hltmp->body++; isspace(*hltmp->body); hltmp->body++);
    }
    
    if (hdo->hlin == NULL)  hdo->hlin = hltmp;
    else  hdo->last->next = hltmp;
    
    hdo->last = hltmp;
    hdo->len += len;
    hdo->num++;
    
    return NULL;
}

void head_clean(Head *hdo)
{
    pm_free(hdo->pm);
    hdo->len = 0;
    hdo->num = 0;
    hdo->fmsp = NULL;
    hdo->hlin = NULL;
    hdo->last = NULL;
    hdo->skey = NULL;
    hdo->scur = NULL;
}

/* put the head part in the file into head cache 
   Note: when call this function, the FromSps line should be inside the
   'buf' argument, and file seeking in the 'fin' has moved to next line */

int head_to_cache(Head *hdo, FILE *fin, char *buf, int len)
{
    hdo->fmsp = pm_dup_str(hdo->pm, buf);
    while (fgets(buf, len, fin))  {	
	if (head_loading(hdo, buf) != NULL)  break;
    }
    return 0;
}

int head_of_file(Head *hdo, FILE *fin, char *buf, int len)
{	
    /* skip leading nonsense lines */

    while (fgets(buf, len, fin))  {
	if (ismailstart(buf))  
	    return head_to_cache(hdo, fin, buf, len);
    }

    return -1;
}   

char *head_append(Head *hdo, char *head)
{
    return head_loading(hdo, head);
}	


void head_remove_first(Head *hdo, char *key)
{
    HLin  *hl;

    for (hl = hdo->hlin; hl; hl = hl->next)  {
	if (!strcasecmp(hl->key, key))  {
	    hl->rm = 1;
	    break;
	}
    }
}

void head_remove_all(Head *hdo, char *key)
{
    HLin  *hl;

    for (hl = hdo->hlin; hl; hl = hl->next)  {
	if (!strcasecmp(hl->key, key))  hl->rm = 1;
    }
}


/* check heads against given mask. If rm set, remove all heads in the mask;
 * otherwise remove all heads except heads in the mask */

void head_mask(Head *hdo, char *mask, int rm)
{
    HLin  *hl;
    char  *p, *k, flag, buf[512];

    strncpy(buf, mask, 511);
    buf[511] =  0;
    for (hl = hdo->hlin; hl; hl = hl->next)  {
	p = k = buf;
	flag = 0;
	while (*p)  {
	    k = strchr(p, ':');
	    if (k)  *k = 0;

	    if (!strcasecmp(hl->key, p))  {
		if (rm)  hl->rm = 1;
		flag = 1;
	    }

	    if (k)  *k = ':', p = k + 1;
	    else  break;
	    if (flag == 1)  break;
	}
	if (!flag && !rm)  hl->rm = 1;
    }
}

    
HLin *head_check(Head *hdo, char *key)
{
    HLin  *hl;

    hdo->skey = pm_dup_str(hdo->pm, key);
    for (hl = hdo->hlin; hl; hl = hl->next)  {
	if (!strcasecmp(hl->key, key) && !hl->rm)  {
	    hdo->scur = hl->next;
	    return hl;
	}
    }
    return NULL;
}


HLin *head_check_next(Head *hdo)
{
    HLin  *hl;

    if (hdo->skey == NULL)  return NULL;
    
    for (hl = hdo->scur; hl; hl = hl->next)  {
	if (!strcasecmp(hl->key, hdo->skey) && !hl->rm)  {
	    hdo->scur = hl->next;
	    return hl;
	}
    }
    return NULL;
}


char *head_fetch(HLin *h)
{
    char  *p;
    struct  head_chain  *hc;

    if ((p = malloc(h->len + 1)) == NULL)  return NULL;

    strcpy(p, h->headin);
    for (hc = h->hc; hc; hc = hc->next)  strcat(p, hc->str);

    return p;
}


char *head_fetch_body(HLin *h)
{
    char  *p;
    struct  head_chain  *hc;

    if ((p = malloc(h->len + 1)) == NULL)  return NULL;

    strcpy(p, h->body);
    for (hc = h->hc; hc; hc = hc->next)  strcat(p, hc->str);

    return p;
}

/* Has the specified class headline existed ? 
   It check the head index (class) first, then search the cache for 
   the same class head */

HLin *head_exist(Head *hdo, char *head)
{
    char  idx[256];
    HLin  *hl;

    if (head_index(head, idx, 256) == NULL)  return NULL;
    
    for (hl = hdo->hlin; hl; hl = hl->next)  {
	if (!strcasecmp(hl->key, idx) && !hl->rm)  return hl;
    }
    return NULL;
}
	    
    
void head_dump(Head *hdo, FILE *fout, int fmsp)
{
    HLin   *hl;
    struct head_chain *hc;

    if (hdo->fmsp && fmsp)  fputs(hdo->fmsp, fout);
    
    for (hl = hdo->hlin; hl; hl = hl->next)  {
	if (hl->rm)  continue;

	if (hl->headin)  fputs(hl->headin, fout);
	
	for (hc = hl->hc; hc; hc = hc->next)  
	    fputs(hc->str, fout);
    }
    fflush(fout);
}


char *head_index(char *headline, char *buf, int len)
{
    char  *p;
    
    strncpy(buf, headline, len);
    buf[len-1] = 0;
    p = strchr(buf, ':');
    if (p == NULL)  return NULL;
    *p = 0;
    chop_tail(buf);
    return buf;
}
    
	
int head_cmp(char *s, char *d)
{
    char  sh[256], dh[256];

    if (!s || !d)  return 1;
    
    s = head_index(s, sh, 256);
    d = head_index(d, dh, 256);
    
    if (!s && !d)  return 0;	/* they are all the same non-heads */
    if (!s || !d)  return 1;	
    
    return strcasecmp(sh, dh);
}

/*
static int isblankline(char *buf)
{
    while (*buf)  {
       if (isspace(*buf))  buf++;
       else  return 0;
    }
    return 1;
}

static int isheadline(char *buf)
{
    if (isblankline(buf))  return 0;
    if (isspace(*buf))  return 1;
    if (strchr(buf, ':'))  return 1;
    return 0;
}
*/


static char *chop_tail(char *p)
{
    int  i;

    if (p == NULL)  return p;
    for (i = strlen(p); (i >= 0) && (p[i] >= 0) && (p[i] <= ' '); p[i--] = 0);
    return p;
}


#ifdef	EXECUTABLE


int main(int argc, char **argv)
{
    FILE  *fp;
    Head  *head;
    HLin  *hl;
    char  buf[512], tmp[512];

    if (argc < 2)  {
	printf("usage:  %s folder\n", argv[0]);
	return 0;
    }

    head = head_register(4096, 128);
    if (head == NULL)  return 1;
    
    fp = fopen(argv[1], "r");
    if (fp == NULL)  return 1;

    while (!feof(fp))  {
	head_of_file(head, fp, buf, 512);

	head_append(head, "myhead1: nothing at all\n");
	head_append(head, "myhead2: good day\n and bad day\n");

	printf("> Head %d items, %d bytes in total.\n", head->num, head->len);
	head_dump(head, stdout, 1);

	while (fgets(tmp, 512, stdin))  {
	    chop_tail(tmp);
	    if ((hl = head_check(head, tmp)) != NULL)  {
		printf("## %s", hl->headin);
		head_remove_all(head, tmp);
		head_dump(head, stdout, 1);
	    }
	}
	
	head_clean(head);
	printf("\n");
    }

    fclose(fp);
    head_destruct(head);
    return 0;
}

#endif


