
/*
   mime.c -- decode message body
   
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>

#include "defs.h"


/* For example:
	------=_NextPart_000_008A_01C0B63E.41DCF080
	Content-Type: text/plain;
        		charset="gb2312"
	Content-Disposition: attachment;
	        filename="E:\dirhtml2m.txt"
	Content-Transfer-Encoding: base64
*/

typedef	struct	{
	char	*name;	    /* filename="E:\dirhtml2m.txt" */
	char	*bndry;	    /* ------=_NextPart_000_008A_01C0B63E.41DCF080 */
	int	type;	    /* text/plain 1:text; 0:saving file */
	int	encode;	    /* base64 */
	int	chset;	    /* "gb2312" 0:ignore; 1: HZ encoded */
	char	buff[256];
} MMARK;

#define	MAXMK	12


static char *get_boundary(char *sour, char *bndry, int len)
{
    char *p;

    p = strchr(sour, '=');
    if (p == NULL)  
	return NULL;

    for (p++; isspace(*p); p++);
    if (*p == '"')  
	p++;

    *bndry++ = '-';
    *bndry++ = '-';
    StrnCpy(bndry, p, len);

    for (p = bndry + strlen(bndry) - 1; p > bndry; p--)  {
	if (isspace(*p) || (*p == '"'))  
	    *p = 0;
	else 
	    break;
    }

    /* printf("BOUNDARY EXTRACT: %s\n", bndry - 2); */
    return bndry - 2;
}


static void parse_ctype(MMARK *mm, char *ctype)
{
    char  *idx[32], *p;
    int   i, rs;

    rs = ziptoken(ctype, idx, 32, ";");
    
    for (i = 0; i < rs; i++)  {
	p = skip_space(idx[i]);
	if (!strncasecmp(p, "text/", 5))  
	    mm->type = 1;
	else if (!strncasecmp(p, "charset", 7))  {
		
	    p += 7;
	    if (isalpha(*p))  
		continue;
	    while (*p && !isalpha(*p))  
		p++;
	    if (!strncasecmp(p, "hz-gb-2312", 10)) 
		mm->chset = 1;
	    
	} else if (!strncasecmp(p, "boundary", 8))  {
		
	    p += 8;
	    if (isalpha(*p))
		continue;
	    mm->bndry = &mm->buff[128];
	    get_boundary(p, mm->bndry, 120);
	    
	} else if (!strncasecmp(p, "name", 4))  {
	    
	    p += 4;
	    if ((p = strchr(p, '=')) == NULL)
		continue;
	    p++;
	    p = skip_space(p);
	    if (*p == '"')
		p++;
	    /* conPrintf(p); */

	    mm->name = mm->buff;
	    strncpy(mm->name, p, 128);
	    for (p = mm->name + strlen(mm->name) - 1; p >= mm->name; p--)  {
		if (!isspace(*p) && (*p != '"'))
		    break;
		else
		    *p = 0;
	    }
	}
    }
}

static void parse_disposit(MMARK *mm, char *disp)
{
    char  *idx[32], *p;
    int   i, rs;

    rs = ziptoken(disp, idx, 32, ";");

    for (i = 0; i < rs; i++)  {
	p = skip_space(idx[i]);
	if (!strncasecmp(p, "filename", 8))  {
		
	    p += 8;
	    if ((p = strchr(p, '=')) == NULL)
		continue;
	    p++;
	    p = skip_space(p);
	    if (*p == '"')
		p++;
	    /* conPrintf(p); */

	    mm->name = mm->buff;
	    strncpy(mm->name, p, 128);
	    for (p = mm->name + strlen(mm->name) - 1; p >= mm->name; p--)  {
		if (!isspace(*p) && (*p != '"'))
		    break;
		else
		    *p = 0;
	    }
	}
    }
}		
		
static void get_mime_mark(FILE *fp, FILE *fout, MMARK *mm, char *buf, int len)
{
    char  *p;
    int   status;

    status = 0;
    memset(mm, 0, sizeof(MMARK));
    
    while (fgets(buf, len, fp))  {
	
	fputs(buf, fout);
	if (isblankline(buf))
	    break;
	
	p = buf;
	if (!isspace(buf[0]))  {
	    status = 0;
	    if (!strncasecmp(buf, "Content-Type:", 13))  {
		status = 1;
		p = buf + 13;
	    } else if (!strncasecmp(buf, "Content-Transfer-Encoding:", 26))  {
		status = 2;
		p = buf + 26;
	    } else if (!strncasecmp(buf, "Content-Disposition:", 20))  {
		status = 3;
		p = buf + 20;
	    }
	}
	p = skip_space(p);
	/* printf("--COntent:> %s", p); */
	
	switch (status)  {
	case 1:
	    parse_ctype(mm, p);
	    break;

	case 2:
	    if (!strncasecmp(p, "7bit", 4))  
		    mm->encode = CTE_7BIT;
	    else if (!strncasecmp(p, "quoted-printable", 16))  
		    mm->encode = CTE_QP;
	    else if (!strncasecmp(p, "base64", 6))  
		    mm->encode = CTE_BASE64;
	    else if (!strncasecmp(p, "8bit", 4))  
		    mm->encode = CTE_8BIT;
	    else if (!strncasecmp(p, "binary", 6))  
		    mm->encode = CTE_BINARY;
	    break;

	case 3:
	    parse_disposit(mm, p);
	    break;
	    
	default:
	    break;
	}
    }
}


static void
mime_extract(FILE *fin, FILE *fout, char *bndry, int ctl, char *buf, int len)
{
    MMARK  mm;
    FILE   *ftmp;
    char   *p;
    
    while (!feof(fin))  {

	fputs(buf, fout);
	chop(buf);
	
	if (strncmp(buf, bndry, strlen(bndry)))  {
	    fgets(buf, len, fin);
	} else {
	    get_mime_mark(fin, fout, &mm, buf, len);
	    
	    /*printf("mime type:   %d\n", mm.type);
	    printf("mime name:   %s\n", mm.name);
	    printf("mime bndry:  %s\n", mm.bndry);
	    printf("mime encode: %d\n", mm.encode);
	    printf("mime chset:  %d\n", mm.encode);*/
	    
	    if (mm.bndry && *mm.bndry)  {
		fgets(buf, len, fin);
		mime_extract(fin, fout, mm.bndry, ctl, buf, len); 
	    } else if (mm.type == 1)  {
		    
		if (mm.encode == CTE_BASE64)  {
		    fputc('\n', fout);
		    debase64_bndry(fin, fout, bndry, buf, len);
		    fputc('\n', fout);
		} else if (mm.encode == CTE_QP)  {
		    fputc('\n', fout);
		    deqp_bndry(fin, fout, bndry, buf, len);
		    fputc('\n', fout);
		} else {
		    fgets(buf, len, fin);
		}
		
	    } else if (mm.name)  {
		    
		conPrintf("Extract file [%s]? (Yes/No/Rename) ",
			mm.name);
		tcflush(0, TCIFLUSH);
		read(0, buf, len);
		p = skip_space(buf);

		switch (*p) {
		case 'y':
		case 'Y':
		    strcpy(buf, mm.name);
		    break;

		case 'r':
		case 'R':
		    conPrintf("Input filename: ");
		    tcflush(0, TCIFLUSH);
		    read(0, buf, len);
		    chop(buf);
		    if (isblankline(buf))  
			strcpy(buf, mm.name);
		    break;

		default:
		case 'n':
		case 'N':
		    fgets(buf, len, fin);
		    continue;
		}
		
		if ((ftmp = fopen(buf, "w")) == NULL)  {
		    fgets(buf, len, fin);
		    continue;
		}
		if (mm.encode == CTE_BASE64) {
		    debase64_bndry(fin, ftmp, bndry, buf, len);
		} else if (mm.encode == CTE_QP) {
		    deqp_bndry(fin, ftmp, bndry, buf, len);
		} else {
		    while (fgets(buf, len, fin))  {
			if (!strncmp(buf, bndry, strlen(buf) - 1))  
			    break;
			fputs(buf, ftmp);
		    }
		}
		fclose(ftmp);
		
	    } else 
		fgets(buf, len, fin);
	}
    }
    conPrintf("Mime message extracted.\n");
}


static int mime_multi(char *file, char *bndry, int ctl, char *buf, int len)
{
    FILE  *fin, *fout;

    if ((fin = fopen(file, "r")) == NULL) 
	return -1;

    unlink(file);
    if ((fout = fopen(file, "w")) == NULL)  {
	fclose(fin);
	return -1;
    }

    fgets(buf, len, fin);
    mime_extract(fin, fout, bndry, ctl, buf, len);

    fclose(fin);
    fclose(fout);

    return 0;
}


static int decode_hzgb(char *file)
{
    FILE  *fin, *fout;

    if ((fin = fopen(file, "r")) == NULL)
	return -1;

    unlink(file);
    if ((fout = fopen(file, "w")) == NULL)  {
	fclose(fin);
	return -1;
    }

    dehz_file(fin, fout);

    fclose(fin);
    fclose(fout);
    return 0;
}


/* ctl=0: decode attached file; ctl=1: decode text only */

int mime_body(Mail *m, FILE *fin, char *outfile, int ctl)
{
    FILE  *fout;
    char  buf[SVRBUF], bndry[128], *idx[32], *p;
    int   i, rs;

    /*
    if (fin == NULL)  {
	fin = fopen(m->msg->name, "r");
	if (fin == NULL)  {
	    dbgPrintf("mime_body: cannot open input file\n");
	    return -1;
	}
	fseek(fin, m->msg->boff, SEEK_SET);
    }*/

    if ((fout = fopen(outfile, "w")) == NULL)  {
	dbgPrintf("mime_body: cannot open output file. [%s]\n", outfile);
	return -1;
    }

    if (m->c_encode == CTE_QP)
	deqp_file(fin, fout);
    else if (m->c_encode == CTE_BASE64)
	debase64_file(fin, fout);
    else
	dup_stream(fin, fout);

    fclose(fout);
    
    if (m->c_type) {
	    
	StrnCpy(buf, m->c_type, SVRBUF);
	rs = ziptoken(buf, idx, 32, ";");
	
	/*
	printf("Type: ");
	for (i = 0; i < rs; i++)  printf("@ %s ", idx[i]);
	printf("\n"); */
	
	for (i = 0; i < rs; i++)  {
		
	    p = skip_space(idx[i]);	/* skip leading white spaces */
	    
	    if (!strncasecmp(p, "charset", 7))  {
		p += 7;
		if (isalpha(*p))
		    continue;	/* avoid of "charsetxxx" */
		
		while (!isalpha(*p))
		    p++;
		if (!strncasecmp(p, "hz-gb-2312", 10))  {
		    decode_hzgb(outfile);
		    break;
		}
	    } else if (!strncasecmp(p, "boundary", 8))  {
		p += 8;
		if (isalpha(*p))
		    continue;

		p = get_boundary(p, bndry, 128);
		
		/* no code now */
		mime_multi(outfile, p, 0, buf, SVRBUF);
		
		break;
	    }
	}
    }

    return 0;
}
		

/******************************************************************/

/*
    abc =?gb2312?B?xxxxxxx?= hello
        | |      | |      | |
	| |      | |      | +--- tail
	| |      | |      +----- bodyend
	| |      | +--- body
	| |      +----- encode
	| +--- codepage
	+----- start point
*/

struct	code_mark  {
	char	*codepage;
	char	*encode;
	char	*body;
	char	*bodyend;
	char	*tail;
};

#define	ISSPACE(p)	(((p) == '\n') || ((p) == '\r') || ((p) == '\f'))

static struct code_mark *check_code_mark(char *sour)
{
    static struct code_mark cm;
    char  *p = sour;

    if ((*p != '=') || (*(p+1) != '?'))
	return NULL;
    cm.codepage = p + 2;
    
    for (p += 2; *p && (*p != '?'); p++)  {
	if (ISSPACE(*p))
	    *p = ' ';
    }
    if (*p == 0) 
	return NULL;
    cm.encode = p + 1;
    
    for (p++; *p && (*p != '?'); p++)  {
	if (ISSPACE(*p))
	    *p = ' ';
    }
    if (*p == 0) 
	return NULL;
    cm.body = p + 1;
    
    for (p++; *p; p++)  {
	if ((*p == '?') && (*(p+1) == '=')) 
	    break;
	if (ISSPACE(*p)) 
	    *p = ' ';
    }
    if (*p == 0) 
	cm.tail = cm.bodyend = p;
    else  {
	cm.bodyend = p;
	cm.tail = p + 2;
    }

    return &cm;
}   

/* de-code all kinds format if I know in just a single comment 
 * and keep newline char out. Such as "abc =?gb2312?B?xxxxxxx?= hello"
 * There are 3 control style in the variety ctl:
 * 0: No decode, just pass though the original head and keep the newline out
 * 1: decode and remain the leading char like "=?gb2312?B?"
 * 2: decode and get rid of the leading char 
 * Note: the result will replace the original string */

char *decode_head(char *sour, int ctl)
{
    struct code_mark *cm;
    char *p, *k, tmp;
    int  cnt, i, hzflag;

    p = k = sour;
    hzflag = 0;
    while (*k)  {
	if (ctl && ((cm = check_code_mark(k)) != NULL)) {
	    if (ctl == 1)  {
		for ( ; k < cm->body; *p++ = *k++);
		k = cm->bodyend;
	    } else
		k = cm->tail;
	    
	    tmp = *cm->bodyend;
	    *cm->bodyend = 0;
	    
	    if (!strncasecmp(cm->codepage, "hz-gb-2312", 10))
		hzflag++;
	    
	    switch (*cm->encode) {
	    case 'b':
	    case 'B':
		/* puts(cm->body); */
		cnt = debase64_str(cm->body, p);
		for (i = 0; i < cnt; i++, p++)  {
		    if (ISSPACE(*p))
			*p = ' ';
		}
		break;

	    case 'q':
	    case 'Q':
		cnt = deqp_str(cm->body, p);
		for (i = 0; i < cnt; i++, p++)  {
		    if (ISSPACE(*p))
			*p = ' ';
		}
		break;

	    default:
		p = k;
	    }
	    *cm->bodyend = tmp;

	} else if (ISSPACE(*k))
	    *p++ = ' ', k++;
	else
	    *p++ = *k++;
    }
    *p = 0;
    
    /* specially encode: HZ-GB code */

    if (hzflag)  {
	/* printf("Sub: %s\n", sour); */
	dehz_str(sour, sour);
    }
	    
    return sour;
}


	
