/*
   smtp.c -- the routine set of SMTP
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "defs.h"


static int smtp_helo(int sock);
static int smtp_from(int sock, char *cmd);
static int smtp_rcpto(int sock, char *cmd);
static int smtp_data(int sock);
static int smtp_dotclose(int sock);
static int smtp_reset(int sock);
static int smtp_quit(int sock);
static int smtp_answer(int sock, char *buf, int len);
static int smtp_transact(int sock, char *cmd, int len);
static int smtp_sendmsg(int sock, Mail *m, char *sender);
	

static int smtp_helo(int sock)
{
    char	buf[SVRBUF];
    int 	rs;
    
    sprintf(buf, "HELO %s", get_user());
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    if (!strncmp(buf, "250", 3))  
	return OK;
    
    return MISANSWER;
}

static int smtp_from(int sock, char *cmd)
{
    char	buf[SVRBUF];
    int 	rs;
    
    if (cmd == NULL)  
	strcpy(buf, "MAIL FROM:<>");
    else  
	sprintf(buf, "MAIL FROM:<%s>", email_pure(cmd));
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    
    if (!strncmp(buf, "250", 3))  
	return OK;
    logPrintf(LOG_INFO, "smtp: %s", buf);
/*    
    strcpy(buf, "QUIT");
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  return rs;
*/
    return MISANSWER;
}

/* this routine has changed. The 'cmd' argument accept several receiption
   address seperated by comma. e.g.: <abc>, <def> ... */

static int smtp_rcpto(int sock, char *cmd)
{
    char  buf[SVRBUF], *p;
    Iadr  *iadr;
    int   cnt = 0;
    
    if ((p = cmd) == NULL) 
	return 0;
    
    while ((iadr = email_addr(p)) != NULL)  {
	if (iadr->rtto)  
	    sprintf(buf, "RCPT TO:%s", iadr->rtto);
	else  
	    sprintf(buf, "RCPT TO:<%s>", iadr->email);
	
	if (smtp_transact(sock, buf, SVRBUF) != OK)  
	    return 0;
		
	if (!strncmp(buf, "250", 3))  
	    cnt++;
	else 
	    logPrintf(LOG_INFO, "smtp: %s", buf);

	if ((p = iadr->next) == NULL)  
	    break;
    }	
    return cnt;
}

static int smtp_data(int sock)
{
    char	buf[SVRBUF];
    int 	rs;
    
    strcpy(buf, "DATA");
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK) 
	return rs;
    
    if (!strncmp(buf, "354", 3))  
	return OK;

    logPrintf(LOG_INFO, "smtp: %s", buf);
    return MISANSWER;
}

static int smtp_dotclose(int sock)
{
    char	buf[SVRBUF];
    int 	rs;
    
    strcpy(buf, ".");
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    
    if (!strncmp(buf, "250", 3))  
	return OK;
    
    logPrintf(LOG_INFO, "smtp: %s", buf);
/*
    strcpy(buf, "QUIT");
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  return rs;
*/
    return MISANSWER;
}

static int smtp_reset(int sock)
{
    char	buf[SVRBUF];
    int 	rs;
    
    strcpy(buf, "RSET");
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    
    return OK;
}

static int smtp_quit(int sock)
{
    char	buf[SVRBUF];
    int 	rs;
    
    strcpy(buf, "QUIT");
    rs = smtp_transact(sock, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    return OK;
}

static int smtp_answer(int sock, char *buf, int len)
{
    char	tmp[SVRBUF];
    int 	rs;
    
    rs = sock_readall(sock, buf, len, timeout);
    if (rs != OK)  
	return rs;
    
    if (verbose)  
	conPrintf("<< %s", buf);
    
    strcpy(tmp, buf);
    while (tmp[3] == '-')  {
    	rs = sock_readall(sock, tmp, SVRBUF, timeout);
    	if (rs != OK)  
	    return rs;
	
    	if (verbose)  
	    conPrintf("<< %s", tmp);
    }
    return OK;
}

static int smtp_transact(int sock, char *cmd, int len)
{
    char  buf[SVRBUF];
    int   rs;
    
    if (verbose)  
	conPrintf("-> %s\n", cmd);
    
    sprintf(buf, "%s\r\n", cmd);
    rs = sock_write(sock, buf, strlen(buf), timeout);
    if (rs != OK) 
	return rs;
    
    rs = smtp_answer(sock, cmd, len);
    return rs;
}

static int smtp_sendmsg(int sock, Mail *m, char *sender)
{
    FILE  *fp;
    char  buf[SVRBUF], *bulk, *tmp;
    int   rs;

    if ((m->behave & ACT_SUSP) != 0)  {
	conPrintf("Suspended message.\n");
	return OK;  /* reserved message */
    }
    
    if ((fp = fopen(m->msg->name, "r")) == NULL)  {
	dbgPrintf("message folder open failed. [%s]\n", m->msg->name);
	return OPENFILERR;
    }
    
    if ((rs = smtp_from(sock, sender)) != OK)  {
	dbgPrintf("reject return path. [%s]\n", sender);
	rs = WARNING;
	goto send_fail;
    }
    
    rs  = smtp_rcpto(sock, m->to);
    rs += smtp_rcpto(sock, m->cc);
    rs += smtp_rcpto(sock, m->bcc);
    if (rs == 0)  {
	dbgPrintf("reject all target addresses, abort it.\n");
	rs = WARNING;
	goto send_fail;
    }

    rs = smtp_data(sock);
    if (rs == MISANSWER)  {
	dbgPrintf("data transfering refused by peer!\n");
	goto send_fail;
    }

    *buf = '.';
    bulk = buf + 1;

    /* skip starting FromSps line */

    while (fgets(bulk, SVRBUF - 1, fp))  {
	if (ismailstart(bulk))
	    break;
    }
    
    prog_init(m->msg->msize - ftell(fp), 40);
    while (fgets(bulk, SVRBUF - 1, fp))  {
	tmp = bulk + strlen(bulk) - 1;
	if (*tmp == '\n')  {
	    *tmp++ = '\r', *tmp++ = '\n', *tmp = '\0';
	}
	
	if (*bulk == '.')  {
	    prog_do(strlen(buf));
	    rs = linesend(sock, buf, timeout);
	} else {
	    prog_do(strlen(bulk));
	    rs = linesend(sock, bulk, timeout);
	}
	
	if (rs != OK)
	    goto send_fail;
    }
    prog_end();
    rs = sflush(sock, timeout);
    if (rs != OK)
	goto send_fail;
    
    rs = smtp_dotclose(sock);
    if (rs != OK)
	goto send_fail;
    
    /* so, the message has sent successfully, mark it as removed. */

    m->removed = 1;
    fclose(fp);
    return OK;
    
send_fail:
    fclose(fp);
    smtp_reset(sock);
    return rs;
}


/* create an smtp server list.
 * The input argument 'host' and 'auth' comes from command line. It returns
 * an smtp server structure for going to use. The step is:
 *  1. First it creates a server chain by the appearence order according to 
 *     the options 'smtp' positing in the config file.
 *  2. If the 'auth' argument is given, it will replace the coresponsive 
 *     elements in the 'smtp' option.
 *  3. If the 'host' argument is given with an Internet address, it will create
 *     a new server structure and then will insert it to the head.
 *  4. If the 'host' argument is given with the form of "@number", such as @2,
 *     it will return the element in the list by the 'number', in this example
 *     it is the second element, so that Pmail will connect the specified smtp
 *     server. Otherwise the first element in the queue will be returned.
 *  5. The 'auth' originally is with the form of 'id:passwd'. But there is an
 *     extensive form available, 'id:passwd:reply-address', for connecting
 *     smtp server only. With another context such as POP3 server it's invalid.
 *  6. If some entry is not written down, that means no need of it. The same
 *     effect can be made by the (*) character. If a blank element is wanted,
 *     you should leave the entry blank. But if your choice is actually the (*)
 *     character, then you should use (\*).
 */

static SMTP *smtpsvr = NULL;
static char *lauth = NULL;
static char *defhost = NULL;
static char *defreply = NULL;

SMTP *smtp_host_chain(char *host, char *auth)
{
    SMTP  *sm = NULL, *tmp;
    char  *idx[4], *auidx[3], *p;
    int   i, rs;
    
    /* first we process the authentication information from command line */
    
    if (auth && (lauth = dup_str(auth)))  {
	rs = fixtoken(lauth, auidx, 3, ":");
	for (i = 0; i < rs; i++)  {
	    if (!strcmp(auidx[i], "*"))
		auidx[i] = NULL;
	    else if (!strcmp(auidx[i], "\\*"))
		auidx[i]++;
	}
    }

    /* then create the smtp server chain */
    
    for (p = rc_check(RCKEY_SMTP); p; p = rc_check_next())  {
	    
	rs = sizeof(SMTP) + strlen(p);
	if ((tmp = malloc(rs)) == NULL)
	    break;
	
	memset(tmp, 0, rs);
	strcpy(tmp->buf, p);
	rs = fixtoken(tmp->buf, idx, 4, ":");
	for (i = 0; i < rs; i++)  {
	    if (!strcmp(idx[i], "*"))
		idx[i] = NULL;
	    else if (!strcmp(idx[i], "\\*"))
		idx[i]++;
	}

	tmp->host = idx[0];
	tmp->reply = idx[1];
	tmp->id = idx[2];
	tmp->passwd = idx[3];
	
	if (lauth)  {
	    tmp->id = auidx[0];
	    tmp->passwd = auidx[1];
	    if (auidx[2])
		tmp->reply = auidx[2];
	}

	if (smtpsvr == NULL)
	    smtpsvr = sm = tmp;
	else {
	    sm->next = tmp;
	    sm = sm->next;
	}   
    }
    
    /* To check which server structure should be returned.
     * If a command line host specified, append an smtp server structure. */
    
    if (host == NULL)
	sm = smtpsvr;
    else  if (*host == '@')  {
	rs = atoi(host+1);
	for (i = 1, sm = smtpsvr; sm; i++, sm = sm->next)  {
	    if (i == rs)
		break;
	}
	if (sm == NULL)
	    sm = smtpsvr;	/* out of smtp server number */
    } else {
	rs = sizeof(SMTP) + strlen(host);
	if ((tmp = malloc(rs)) != NULL) {
	    memset(tmp, 0, rs);
	    strcpy(tmp->buf, host);
	    tmp->host = tmp->buf;
	    if (lauth)  {
		tmp->id = auidx[0];
		tmp->passwd = auidx[1];
		tmp->reply = auidx[2];
	    }

	    tmp->next = smtpsvr;
	    smtpsvr = tmp;
	    
	}
	sm = smtpsvr;
    }

    /* double check to prevent the appearance of invaild elements */

    if (sm == NULL)  {
	sm = smtpsvr = malloc(sizeof(SMTP));
	if (smtpsvr == NULL)
	    return NULL;
	else
	    memset(smtpsvr, 0, sizeof(SMTP));
    }
    if ((sm->host == NULL) || (*sm->host == 0))  {
	defhost = dup_str(get_hostname());
	sm->host = defhost;
    }
    if ((sm->reply == NULL) || (*sm->reply == 0))  {
	p = rc_check(RCKEY_REPLY);
	if (p && *p)  defreply = dup_str(p);
	else {
	    rs = strlen(get_user()) + strlen(get_hostname()) + 5;
	    if ((defreply = malloc(rs)) != NULL)  {
		strcpy(defreply, get_user());
		strcat(defreply, "@");
		strcat(defreply, get_hostname());
	    }
	}
	sm->reply = defreply;
    }
    
    return sm;
}
	
void smtp_host_list()
{
    SMTP *sm;
    char buf[SVRBUF];
    int  i;

    conPrintf("SMTP servers list:\n");
    for (i = 1, sm = smtpsvr; sm; i++, sm = sm->next)  {
	ssprintf(buf, SVRBUF, "%4d %-25s %-25s %s:%s\n", 
			i, sm->host, sm->reply,	sm->id, sm->passwd);
	conPrintf(buf);
    }
    conPrintf("\n");
}

void smtp_host_clean()
{
    SMTP *sm, *tmp;

    if (smtpsvr)  {
	sm = smtpsvr;
	while (sm)  {
	    tmp = sm;
	    sm = sm->next;
	    free(tmp);
	}
    }
    if (lauth)
	free(lauth);
    if (defhost)
	free(defhost);
    if (defreply)
	free(defreply);
}

int smtp_main(Folder *rootfld, SMTP *sm)
{
    Folder *fo;
    Mail   *m;
    char   buf[SVRBUF];
    int    sock, scnt, rs;
    
    
    if (sm == NULL)  {
	logPrintf(LOG_WARN, "No SMTP server existed.\n");
	return FAILURE;
    }

    /* double check the existance of message for sending */
    
    for (rs = 0, fo = rootfld; fo; fo = fo->next)
	rs += fo->num;
    if (rs == 0)  {
	conPrintf("No message to deliver.\n");
	return WARNING;
    }
    
    /* open socket and connect to the destination server */
    
    conPrintf("Trying SMTP server [%s] ... ", sm->host);
    if ((sock = open_socket(sm->host, SMTP_PORT)) < 0)  {
	conPrintf("failed!\n");
	logPrintf(LOG_ERR, "smtp: fail to connect to %s\n", sm->host);
	return sock;
    }
    conPrintf("ok.\n");
    logPrintf(LOG_INFO, "smtp: %s connected for sending.\n", sm->host);
    
    rs = smtp_answer(sock, buf, SVRBUF);
    if (rs != OK)
	return rs;
    if (strncmp(buf, "220", 3))
	return MISANSWER;
    
    rs = smtp_helo(sock);
    if (rs != OK)
	return rs;
    
    /* if ((rs = smtp_reset(sock)) != OK)
	return rs; */
    	
    for (scnt = 0, fo = rootfld; fo; fo = fo->next)  {
	conPrintf("Folder %s, %d messages are ready.\n", fo->name, fo->num);
	for (m = fo->mail; m; m = m->next)  {
	    if (verbose)  {
		conPrintf("sending %3d/%d, (%-8ld)\n", 
			m->id, fo->num, m->msg->msize);
	    } else {
		conPrintf("sending %3d/%d, (%-8ld) ",
			m->id, fo->num, m->msg->msize);
	    }
	    rs = smtp_sendmsg(sock, m, sm->reply);
	    if (rs == OK)
		scnt++;
	    else if (rs != WARNING)
		break;	/* fatal error encountered */
	}
    }
	
    smtp_quit(sock);
    close(sock);
    logPrintf(LOG_INFO, "smtp: %d mails delivered.\n", scnt);

    return rs;
}


