
/*
   strdate.c -- change date string to (struct tm)
	
   Copyright (C) 1998, 2000  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if TIME_WITH_SYS_TIME
  #include <sys/time.h>
  #include <time.h>
#else
  #if HAVE_SYS_TIME_H
    #include <sys/time.h>
  #else
    #include <time.h>
  #endif
#endif


extern int ziptoken(char *sour, char **idx, int ids, char *delim);

static char *week[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
static char *mon[] = { "Jan", "Feb", "Mar",  "Apr", "May", "Jun",
                       "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static int get_week(char *s)
{
    int  i;

    for (i = 0; i < 7; i++)  if (!strncmp(week[i], s, 3))  break;
    if (i == 7)  return -1;
    return i;
}

static int get_month(char *s)
{
    int  i;

    for (i = 0; i < 12; i++)  if (!strncmp(mon[i], s, 3))  break;
    if (i == 12)  return -1;
    return i;
}

static int get_mday(char *s)
{
    int  i;

    i = atoi(s);
    if ((i < 1) || (i > 31))  return -1;
    return i;
}

static int get_year(char *s)
{
    int  i;

    /* The number of years since 1900 */
    
    i = atoi(s);
    
#ifdef	Y2K_FORCED
    if ((i >= 0) && (i < 70))  i += 2000;	/* for Y2K problem */
    else if ((i > 70) && (i < 100))  i+= 1900;
    else if (i > 1900)  i -= 1900;
#else
    if (i > 1900)  i -= 1900;
#endif
    
    return i;
}

/* 
    Time zone may be indicated in several ways.  "UT" is Universal  
    Time  (formerly called "Greenwich Mean Time"); "GMT" is permitted 
    as a reference to Universal Time.  The  military  standard uses  
    a  single  character for each zone.  "Z" is Universal Time. "A" 
    indicates one hour earlier, and "M" indicates 12  hours  earlier;  
    "N"  is  one  hour  later, and "Y" is 12 hours later.  The letter 
    "J" is not used.  The other remaining two forms are  taken from 
    ANSI standard X3.51-1975.  One allows explicit indication of the 
    amount of offset from UT; the other uses  common  3-character 
    strings for indicating time zones in North America.
*/

static int get_zone(char *s)
{
    int  z;

    if (!strcmp(s, "UT") || !strcmp(s, "GMT") || !strcmp(s, "Z"))  z = 0;
    else if (!strcmp(s, "EST"))  z = 500;
    else if (!strcmp(s, "EDT"))  z = 400;
    else if (!strcmp(s, "CST"))  z = 600;
    else if (!strcmp(s, "CDT"))  z = 500;
    else if (!strcmp(s, "MST"))  z = 700;
    else if (!strcmp(s, "MDT"))  z = 600;
    else if (!strcmp(s, "PST"))  z = 800;
    else if (!strcmp(s, "PDT"))  z = 700;
    else if (!strcmp(s, "A"))  z = -100;
    else if (!strcmp(s, "M"))  z = -1200;
    else if (!strcmp(s, "N"))  z = 100;
    else if (!strcmp(s, "Y"))  z = 1200;
    else if ((*s == '+') || (*s == '-'))  z = atoi(s);
    else z = 10000;	/* mark as parse error */

    return z;
}

static int get_time(char *s, struct tm *t)
{
    char  *idx[3];

    if (ziptoken(s, idx, 3, ":") < 2)  return -1;
    if (idx[0])  t->tm_hour = atoi(idx[0]);
    if (idx[1])  t->tm_min  = atoi(idx[1]);
    if (idx[2])  t->tm_sec  = atoi(idx[2]);
    return 0;
}

static int get_mday_rfc850(char *s, struct tm *t)
{
    char  *idx[3];

    ziptoken(s, idx, 3, "-");
    if (idx[0])  t->tm_mday = atoi(idx[0]);
    if (idx[2])  t->tm_year = atoi(idx[2]);
    if ((t->tm_mon = get_month(idx[1])) == -1)  return -1;
    return 0;
}

int strdate(char *s, struct tm *t)
{
    char  *idx[10], buf[80];
    int   zone, rs = 0;

    strncpy(buf, s, 80);
    buf[79] = '\0';
    memset(t, 0, sizeof(struct tm));
    rs = ziptoken(buf, idx, 10, ", ");
    
    if (strchr(s, '-') && (rs == 4))  {

	/* rfc850 date style */
	/* Weekday, DD-Mon-YY HH:MM:SS TIMEZONE */
	/* Is this style Y2K compalient ? */

	rs = 0;
	if ((t->tm_wday = get_week(idx[0])) == -1)  rs++;
	if (get_mday_rfc850(idx[1], t) == -1)  rs++;
	if (get_time(idx[2], t) == -1)  rs++;
	if ((zone = get_zone(idx[3])) == 10000)  rs++;

    } else if (s[3] == ',')  {
	    
	/* rfc822 date style */
	/* [WDY,] DD Mon YY HH:MM[:SS] TIMEZONE */

	rs = 0;
	if ((t->tm_wday = get_week(idx[0])) == -1)  rs++;
	if ((t->tm_mday = get_mday(idx[1])) == -1)  rs++;
	if ((t->tm_mon = get_month(idx[2])) == -1)  rs++;
	if ((t->tm_year = get_year(idx[3])) == -1)  rs++;
	if (get_time(idx[4], t) == -1)  rs++;
	if ((zone = get_zone(idx[5])) == 10000)  rs++;
	
    } else if (*idx[0] <= '9')  {
	
	/* rfc822 date style */
	/* [WDY,] DD Mon YY HH:MM[:SS] TIMEZONE */

	rs = 0;
	if ((t->tm_mday = get_mday(idx[0])) == -1)  rs++;
	if ((t->tm_mon = get_month(idx[1])) == -1)  rs++;
	if ((t->tm_year = get_year(idx[2])) == -1)  rs++;
	if (get_time(idx[3], t) == -1)  rs++;
	if ((zone = get_zone(idx[4])) == 10000)  rs++;

    } else  if ((rs == 5) || (rs == 6)) {

	/* ctime date style */
	/* Note in particular that ctime format:

		Wdy Mon DD HH:MM:SS [TIMEZONE] YYYY

	   is not acceptable because it is not a valid ARPANET  date.
	   However, since older software still generates this format,
	   news implementations are encouraged to accept this  format
	   and translate it into an acceptable format. */
	    
	rs = 0;
	if ((t->tm_wday = get_week(idx[0])) == -1)  rs++;
	if ((t->tm_mon = get_month(idx[1])) == -1)  rs++;
	if ((t->tm_mday = get_mday(idx[2])) == -1)  rs++;
	if (get_time(idx[3], t) == -1)  rs++;

	if (*idx[4] > '9')  {
	    if ((zone = get_zone(idx[4])) == 10000)  rs++;
	    if ((t->tm_year = get_year(idx[5])) == -1)  rs++;
	} else if ((t->tm_year = get_year(idx[4])) == -1)  rs++;
    }
    
    return rs;
}


time_t strtimet(char *s)
{
    struct tm t;

    strdate(s, &t);
    return mktime(&t);
}   
    

#ifdef	EXECUTABLE

int main(int argc, char **argv)
{
    struct tm	t;
    int  rs = 0;
    
    if (argc < 2)  {
	printf("usage: %s \"date string\"\n", argv[0]);
	return 0;
    }

    if ((rs = strdate(argv[1], &t)) > 0) {
	printf("parse error\n");
    } else {
	printf("%s\n", asctime(&t));
    }
    return rs;
}

#endif

